///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoAdmin.cc
// -------------
// Cego admin console program
//
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000-2010 Bjoern Lemke                        
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: Main
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

#include <lfc/Chain.h>
#include <lfc/Exception.h>
#include <lfc/GetLongOpt.h>
#include <lfc/File.h>
#include <lfc/Timer.h>
#include <lfc/Tokenizer.h>
#include <lfc/Net.h>

#include <lfc/Element.h>
#include <lfc/Document.h>
#include <lfc/XMLSuite.h>

#include "CegoDefs.h"
#include "CegoAdmAction.h"
#include "CegoAdmMon.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <readline/readline.h>
#include <readline/history.h>


#define USAGE "Usage: cgadm --user=<user>/<password>\n\
          [ --server=<host>]\n\
          [ --port=<port> ]\n\
          [ --batchfile=<batchfile> ]\n\
          [ --cmd=<command> ]\n\
          [ --logfile=<logfile> ]\n\
          [ --raw ]\n\
          [ --mon ]\n\
          [ --ignore ]\n\
          [ --debug ] [ --version  ] [ --help ]"


#define DEFAULTSERVER "localhost"
#define DEFAULTPORT 2000

#define HISTFILE ".cgadm_history"
#define PROMPT "CGADM > "
#define PROMPT2 "> "

enum RunMode { INTERACTIVE, BATCH, COMMAND, MONITOR };

bool dispatchCmd(CegoAdmAction* pAdm, const Chain& cmd);
void showMonitor(CegoAdminHandler *pAH);

void printHelp(const Chain& cs);
void helpAll();
void helpShow(const Chain& c);
void helpThreadInfo(const Chain& c);
void helpList(const Chain& arg);
void helpCreate();
void helpDefine();
void helpStart();
void helpCopy();
void helpSwitch();
void helpStop();
void helpDrop();
void helpAdd(const Chain& arg);
void helpRemove(const Chain& arg);
void helpPasswd(const Chain& arg);
void helpSync(const Chain& arg);
void helpAbort(const Chain& arg);
void helpImport(const Chain& arg);
void helpExport(const Chain& arg);
void helpBegin();
void helpEnd();
void helpRecover();
void helpEnable();
void helpDisable();
void helpRelocate();
void helpCheck();


extern char __lfcVersionString[];
extern char __lfcxmlVersionString[];

int main(int argc, char **argv)
{	    
    GetLongOpt longOpt(argc, argv);
    
    longOpt.addOpt("version");
    longOpt.addOpt("help");
    longOpt.addOpt("logfile");
    longOpt.addOpt("user");
    longOpt.addOpt("cmd");
    longOpt.addOpt("batchfile");
    longOpt.addOpt("debug");
    longOpt.addOpt("raw");
    longOpt.addOpt("mon");
    longOpt.addOpt("ignore");
    longOpt.addOpt("server", DEFAULTSERVER);
    longOpt.addOpt("port", Chain(DEFAULTPORT));
    longOpt.addOpt("simulate");

    try
    {
	longOpt.parseOpt(); 
    }
    catch ( Exception e )
    {
	Chain msg;
	e.pop(msg);
	cerr << msg << endl;
	cerr << USAGE << endl;
	exit(1);	
    }
    
    Chain user;
    Chain password;
    

#ifdef HAVE_MINGW32
    Chain homeDir(getenv("HOMEPATH"));
    Chain histPath = homeDir + "\\" + Chain(HISTFILE);
#else
    Chain homeDir(getenv("HOME"));
    Chain histPath = homeDir + "/" + Chain(HISTFILE);
#endif
    
    RunMode runMode = INTERACTIVE;

    bool debug=false;
    if ( longOpt.isSet("debug") )
	debug=true;

    bool rawMode = false;
    if ( longOpt.isSet("raw") )
	rawMode=true;

    bool ignoreError = false;
    if ( longOpt.isSet("ignore") )
	ignoreError=true;
    
    if ( longOpt.isSet( Chain("help") ) )
    {	
	cerr << USAGE << endl;
	exit(0);
    }

    if ( longOpt.isSet( Chain("version") ) )
    {
	cout << CEGO_PRODUCT << " (" << sizeof(long) * 8 << " bit), Version " << CEGO_VERSION 
	     << " [ lfc : " << __lfcVersionString  
	     << ", lfcxml : " <<  __lfcxmlVersionString << " ]" << endl;
	cout << CEGO_COPYRIGHT << endl;
	exit(0);
    }

    Chain logFile = longOpt.getOptValue("logfile");
    Chain serverName = longOpt.getOptValue("server");
    int portNo = longOpt.getOptValue("port").asInteger();

    Chain batchFileName;
    Chain command;

    if ( longOpt.isSet("batchfile") )
    {
	runMode = BATCH;
	batchFileName = longOpt.getOptValue("batchfile");
    }
    else if ( longOpt.isSet("cmd") )
    {
	runMode = COMMAND;
	command = longOpt.getOptValue("cmd");
    }

    else if ( longOpt.isSet("mon") )
    {
	runMode = MONITOR;
    }

    CegoModule *pModule = 0;
    CegoAdminHandler* pAH = 0;
    NetHandler* pN = 0;
    unsigned long modId = 0;

    try
    {
	Chain authString = longOpt.getOptValue("user");
	
	Tokenizer authTok(authString, Chain("/")); 
	authTok.nextToken(user);
	authTok.nextToken(password);
	
	if ( user.length() == 0 )
	    throw Exception(EXLOC, "User not set");
	if ( password.length() == 0 )
	    throw Exception(EXLOC, "Password not set");
	
       
	if ( logFile.length() > 0 )
	    pModule = new CegoModule(logFile);
	else
	    pModule = new CegoModule();
		
	modId = pModule->getModId("CegoAdmin");
	
	if ( debug )
	    pModule->logModule(modId, "CegoAdmin", Logger::DEBUG);
	else
	    pModule->logModule(modId, "CegoAdmin", Logger::NOTICE);
		
	Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	
	pN = n.connect(serverName, portNo);
	pAH = new CegoAdminHandler( pModule, pN);
	
	CegoAdminHandler::ResultType res = pAH->requestSession(user, password);
	if ( res != CegoAdminHandler::ADM_OK )
	{
	    Chain msg;
	    pAH->getMsg(msg);
	    throw Exception(EXLOC, msg);
	}
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);
	cerr << msg << endl;
	exit(1);
    }

    pModule->log(modId, Logger::LOGERR, Chain("Got session... "));

    CegoAdmAction* pAdm = new CegoAdmAction(pAH, pModule, rawMode);
    
    int exitCode = 0;

    try
    {

	switch ( runMode )
	{
	case MONITOR:
	{	    
	    CegoAdmMon mon(pAH);
	    mon.showMonitor();
	    break;	    	    
	}
	case BATCH:
	{
	    
	    int lineNo=0;
	    
	    
	    File batchFile(batchFileName);
	    
	    batchFile.open(File::READ);	    
	    
	    Chain cmd;
	    Chain line;
	    bool disableDelimiter=false;
	    bool doQuit = false;
	    
	    while (doQuit == false && batchFile.readLine(line))
	    {
		lineNo++;
		line = line.cutTrailing(" \t");
		
		bool isCommentLine = false;
		
		if ( line.length() >= 2 )
		{		    
		    if (line.cutTrailing(" \t").subChain(1, 2) == Chain("--"))
		    {
			isCommentLine = true;
		    }
		}
		if ( ! isCommentLine )
		{
		    
		    cmd = cmd + Chain(" ") + line;
		    cmd = cmd.cutTrailing(" \t");
		    
		    if ( cmd.length() > 0 )
		    {
			
			if (cmd.subChain(cmd.length()-1, cmd.length()) == Chain(";")) 
			{			
			    try
			    {
				doQuit = dispatchCmd(pAdm, cmd);
			    }
			    catch ( Exception e )
			    {
				cerr << "Line Number " + Chain(lineNo) + " : " << endl;
				Chain msg;
				e.pop(msg);
				cerr << msg << endl;
				exitCode = 1;
				
				pAdm->cleanUp();
				
				if ( ignoreError == true )
				    cerr << "Ignored" << endl;
				else
				    throw Exception(EXLOC, msg);
				
			    }
			    cmd = Chain("");			    
			}
		    }
		}
	    }
	    if ( cmd.length() > 1 )
	    {
		cout << "incomplete command ..<" << cmd << ">"  << endl;
	    }
	    
	    batchFile.close();
	    
	    break;
	}
	case COMMAND:
	{		    
	    
	    Chain cmd = command.cutTrailing(" ;") + Chain(";");
	    dispatchCmd(pAdm, cmd);	    
	    break;
	}
	case INTERACTIVE:
	{
	    	    	
	    read_history((char*)histPath);
	    
	    bool doQuit = false;
	    
	    while ( doQuit == false )
	    {			    		
		
		char *pC = 0;
		
		bool isTerminated = false;
		Chain cs;
		
		pC = readline(PROMPT);
		
		if (pC)
		{
		    cs = Chain(pC);
		    
		    while (! isTerminated )
		    {		
			add_history(pC);	
			
			if (pC)
			    free(pC);	    
			
			cs = cs.cutTrailing(Chain(" \t"));
			
			if ( cs.length() > 1)
			{		
			    
			    if ( cs[cs.length()-2] == ';')
			    {
				
				try
				{
				    doQuit = dispatchCmd(pAdm, cs);
				}
				catch ( Exception e )
				{
				    Chain msg;
				    e.pop(msg);
				    cerr << "ERROR : " << msg << endl;
				}
				
				pAdm->cleanUp();
				
				write_history((char*)histPath);
				
				isTerminated = true;
			    }
			    else if ( cs.length() > 4 )
			    {
				if ( cs.subChain(1,4) == Chain("help") )
				{
				    printHelp(cs);
				    isTerminated = true;
				}
			    }
			    
			}
			
			if ( ! isTerminated )
			{
			    pC = readline(PROMPT2);
			    
			    if (pC)
			    {
				cs = cs + Chain(" ") + Chain(pC);
			    }
			}
		    }		  
		}
		else
		{
		    cout << "Bye" << endl;
		    doQuit=true;
		}
	    }    
	    break;
	}
	}
    }    
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);
	cerr << "ERROR : " << msg << endl;
	exit(1);
    }
    
    pAH->closeSession();
    
    delete pAdm;
    delete pAH;
    delete pN;
    delete pModule;

    exit(exitCode);
}

bool dispatchCmd(CegoAdmAction *pAdm, const Chain& cmd)
{
    
    if ( cmd.cutTrailing(" ;") == Chain("quit") )
	return true;

    pAdm->setCommandChain( (char*)cmd);
    
    Timer t(6,3);
    t.start();
    		    		    
    pAdm->parse();
    
    t.stop();
    if ( pAdm->getRawMode() == false )
	cout << "ok ( " << t << " s )" << endl;

    return false;
}

void printHelp(const Chain &cs)
{
    Tokenizer t(cs, Chain(" "));
    Chain a, b, c;
    t.nextToken(a);
    t.nextToken(b);
    t.nextToken(c);
    
    if ( b == Chain("help"))
    {
	helpAll();
    }
    else if ( b == Chain("show"))
    {
	helpShow(c);
    }
    else if ( b == Chain("threadinfo"))
    {
	helpThreadInfo(c);	
    }
    else if ( b == Chain("list"))
    {
	helpList(c);
    }
    else if ( b == Chain("define"))
    {
	helpDefine();
    }
    else if ( b == Chain("create"))
    {
	helpCreate();
    }
    else if ( b == Chain("start"))
    {
	helpStart();
    }
    else if ( b == Chain("copy"))
    {
	helpCopy();
    }
    else if ( b == Chain("switch"))
    {
	helpSwitch();
    }
    else if ( b == Chain("stop"))
    {
	helpStop();
    }
    else if ( b == Chain("drop"))
    {
	helpDrop();
    }
    else if ( b == Chain("remove"))
    {
	helpRemove(c);
    }
    else if ( b == Chain("sync"))
    {
	helpSync(c);
    }
    else if ( b == Chain("export"))
    {
	helpExport(c);
    }
    else if ( b == Chain("import"))
    {
	helpImport(c);
    }
    else if ( b == Chain("add"))
    {
	helpAdd(c);
    }
    else if ( b == Chain("passwd"))
    {
	helpPasswd(c);
    }
    else if ( b == Chain("abort"))
    {
	helpAbort(c);
    }
    else if ( b == Chain("begin"))
    {
	helpBegin();
    }
    else if ( b == Chain("end"))
    {
	helpEnd();
    }
    else
    {
	cout << endl;
	cout << "Available Commands" << endl;	
	cout << "==================" << endl;
	cout << endl;
	cout << "   Format commands" << endl;
	cout << "   ---------------" << endl;
	cout << "   rawmode on" << endl;
	cout << "   rawmode off" << endl;
	cout << endl;
	cout << "   Database information" << endl;
	cout << "   --------------------" << endl;
	cout << endl;
	cout << "   show pool" << endl;
	cout << "   list pool" << endl;
	cout << "   list < rec | rbrec | sysrec | bp | dp | ix | sp | df | tp | ts > lock" << endl;
	cout << "   lockstat" << endl;
	cout << "   list tableset" << endl;
	cout << "   list datafile for <tableset>" << endl;
	cout << "   list archlog for <tableset>" << endl;
	cout << "   list logfile for <tableset>" << endl;
	cout << "   list transaction for <tableset>" << endl;
	cout << "   list object for <tableset>" << endl;
	cout << "   list node" << endl;
	cout << "   list session" << endl;
	cout << "   list copy" << endl;
	cout << "   threadinfo [ db [threadid] | adm | log ]" << endl;
	cout << "   abort db thread <tid>" << endl;

	cout << endl;
	cout << "   Tableset administration" << endl;
	cout << "   -----------------------" << endl;
	cout << endl;
	cout << "   retrieve tableset from '<hostname>'" << endl;
	cout << "   define tableset <tableset> <attr1>=<value1> <attr2=value2> ..." << endl;	
	cout << "   create tableset" << endl;
	cout << "   info tableset <tableset>" << endl;
	cout << "   start tableset <tableset>" << endl;
	cout << "   stop tableset <tableset>" << endl;
	cout << "   copy tableset <tableset>" << endl;
	cout << "   switch tableset <tableset>" << endl;
	cout << "   switch secondary for <tableset>" << endl;
	cout << "   switch mediator for <tableset>" << endl;
	cout << "   relocate secondary for <tableset> to '<hostname>'" << endl;
	cout << "   relocate mediator for <tableset> to '<hostname>'" << endl;
	cout << "   drop tableset <tableset>" << endl;
	cout << "   remove tableset <tableset>" << endl;
	cout << "   add <app|sys|temp> datafile '<filename>' size <numpages> to <tableset>" << endl;
	cout << "   check tableset <tableset>" << endl;
	cout << "   verify tableset <tableset>" << endl;
	cout << "   checkpoint interval <intval> for <tableset>" << endl;
	cout << endl;
	cout << "   User administration" << endl;
	cout << "   -------------------" << endl;
	cout << endl;
	cout << "   add user <userid> identified by '<passwd>' to <tableset>" << endl;
	cout << "   remove user <userid> from <tableset>" << endl;
	cout << "   list user for <tableset>" << endl;
	cout << "   passwd user <userid> identified by '<passwd>' to <tableset>" << endl;
	cout << "   add role <role> to user <userid> for <tableset>" << endl;
	cout << "   remove role <role> from user <userid> for <tableset>" << endl;
	cout << "   trace [ on | off ] user <userid> for <tableset>" << endl;
	cout << endl;
	cout << "   Backup and Recovery" << endl;
	cout << "   -------------------" << endl;
	cout << "   set logmng '<progname>'" << endl;
	cout << "   enable archlog for <tableset>" << endl;
	cout << "   disable archlog for <tableset>" << endl;
	cout << "   add archlog <archid> path '<archpath>' to <tableset>" << endl;
	cout << "   remove archlog <archid> from <tableset>" << endl;
	cout << "   begin backup for <tableset> [ keep ticket ]" << endl;
	cout << "   end backup for <tableset>" << endl;
	cout << "   sync tableset <tableset> [ with '<command>' [ <timeout> ]]] " << endl;
	cout << "   recover tableset <tableset> [ to <point in time> ]" << endl;
	cout << endl;
	cout << "   [ xml | binary | plain ] export table <table> from <tableset> to '<filename>'" << endl;
	cout << "   [ xml | binary | plain ] export tableset [ structure ] <tableset> to '<filename>'" << endl;
	cout << endl;
	cout << "   [ xml | binary | plain ] import table <table> to <tableset> from '<filename>' [ nologging ]" << endl;
	cout << "   [ xml | binary | plain ] import tableset [ structure ] <tableset> from '<filename>' [ nologging ]" << endl;
	cout << endl;
	cout << "Close each command with <;> ( e.g.: 'poolinfo;' )" << endl;
	cout << endl;
	cout << "Type \"help help\" ( without ';' ) for online help" << endl;
	
    }
}

void helpAll()
{

    cout << "Type \"help <topic>\" for a detailed help" << endl;
    cout << "Topic is one of the following items" << endl;
    cout << endl;
    
    cout << "   show" << endl;
    cout << "   threadinfo" << endl;
    cout << "   list" << endl;
    cout << "   define" << endl;
    cout << "   create" << endl;
    cout << "   start" << endl;
    cout << "   copy" << endl;
    cout << "   switch" << endl;
    cout << "   stop" << endl;
    cout << "   drop" << endl;
    cout << "   remove" << endl;
    cout << "   sync" << endl;
    cout << "   import" << endl;
    cout << "   export" << endl;
    cout << "   add" << endl;
    cout << "   passwd" << endl;	
    cout << "   abort" << endl;
    cout << "   begin" << endl;
    cout << "   end" << endl;
    cout << "   recover" << endl;
    cout << "   enable" << endl;
    cout << "   disable" << endl;
    cout << "   relocate" << endl;
}

void helpShow(const Chain& arg)
{

    if ( arg == Chain("pool") )
    {
	cout << endl;
	cout << "show pool" << endl;
	cout << "---------" << endl;
	cout << endl;
	cout << "   The poolinfo lists the current status of the database buffer cache parameters." << endl;
	cout << endl;
	cout << "   This information may be helpful for performance optimization ( e.g. increase the buffer cache size)" << endl;
	cout << endl;
	cout << "   The following parameters are indicated : " << endl;
	cout << endl;
	cout << "   Page Size         - The size of each page in the buffercache." << endl;
	cout << "                       This parameter is specified at database creation time and can not be changed." << endl;
	cout << "                       All subsequent information is given in units of pages based on the given pagesize." << endl;
	cout << endl;
	cout << "   Total Pages       - The number of pages available in the bufferpool." << endl;
	cout << "                       There are no special regions in the buffercache for ordering, rollback segments, indices and so on." << endl;
	cout << "                       Page requests for any reasons are assigned using a modulo hash algorithm" << endl;
	cout << "                       by treating the corresponding fileId and pageId of the managed database file" << endl;
	cout << endl;
	cout << "   Used Pages        - The number of pages which are in use. Actual database information is cached in this pages." << endl;
	cout << endl;
	cout << "   Free Pages        - Pages which are still available and have not been filled with any valid data" << endl;
	cout << endl;
	cout << "   Dirty Pages       - Pages filled with valid and modified data." << endl;
	cout << endl;
	cout << "   Fixed Pages       - Pages, which are currently in use by any transaction." << endl;
	cout << endl;
	cout << "   Persistent Pages  - Pages, which can not be removed from the buffer pool."<< endl;
	cout << "                       Normally, thes pages contain database dictionary information" << endl;
	cout << endl;
	cout << "   No Sync Pages     - Forperformance reasones, sometimes pages are used which are not in sync with the transaction log." << endl;
	cout << "                       These  pages can be written to disk as required." << endl;
	cout << "                       E.g. for creation of indexes, no sync pages are used." << endl;
	cout << endl;
	cout << "   Stat Start        - Start point of statistics measurement. The following indicated values are collected since this time" << endl;
	cout << endl;
	cout << "   Disk Reads        - Performed disk database block read opertions since start of statistics measurement." << endl;
	cout << endl;
	cout << "   Disk Writes       - Performed disk database block write opertions since start of statistics measurement." << endl;
	cout << endl;
	cout << "   Read Delay        - Average delay of a disk database block read operation, measured since start of statistics." << endl;
	cout << endl;
	cout << "   Write Delay       - Average delay of a disk database block write operation, measured since start of statistics." << endl;
	cout << endl;
    }
    else if ( arg == Chain("user") )
    {
	cout << endl;
	cout << "show user for <tableset>" << endl;
	cout << "------------------------" << endl;
	cout << endl;
	cout << "   Lists the user of the given tableset." << endl;
	cout << endl;	
    }
    else
    {
	cout << "Available help for show is" << endl;
	cout << endl;
	cout << "   help show pool" << endl;
	cout << endl;
    }
}

void helpThreadInfo(const Chain& arg)
{

    if ( arg == Chain("db") )
    {
    }
    else if ( arg == Chain("adm") )
    {
    }
    else if ( arg == Chain("log") )
    {
	
    }
    else 
    { 
	cout << endl;
	cout << "threadinfo [ db | adm | log]" << endl;
	cout << "------------------------------" << endl;
	cout << endl;
	cout << "   With arugument, the threadinfo command lists all configured and current active threads in the database system." << endl;
	cout << "   The information is given for all three thread pools ( db, adm and log )." << endl;
	cout << endl;
	cout << "   More detailed information for each thread pool is given, if a pool id is specified with the command." << endl;
	cout << "   Valid pools id's are db, adm or log." << endl;
	cout << endl;
    }
}

void helpList(const Chain& arg)
{
    if ( arg ==  Chain("tableset"))
    {
	cout << endl;
	cout << "list tableset" << endl;
	cout << "-------------" << endl;
	cout << endl;
	cout << "   List all known tablesets on the node." << endl;
	cout << "   The current status and assigned hosts ( mediator, primary, secondary) are also indicated." << endl;
	cout << endl;	
    }
    else if ( arg ==  Chain("datafile"))
    {
	cout << endl;
	cout << "list datafile" << endl;
	cout << "-------------" << endl;
	cout << "   List all configured datafiles for the given tableset" << endl;
	cout << endl;	
    }
    else if ( arg ==  Chain("archlog"))
    {
	cout << endl;
	cout << "list archlog" << endl;
	cout << "-----------" << endl;
	cout << "   List all configured archive log locations for the given tableset" << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("user"))
    {
	cout << endl;
	cout << "list user for <tableset>" << endl;
	cout << "------------------------" << endl;
	cout << "   List all defined user for the given tableset" << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("lock"))
    {
	cout << endl;
	cout << "list lock" << endl;
	cout << "------------------------" << endl;
	cout << "   List all lock objects." << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("node"))
    {
	cout << endl;
	cout << "list node" << endl;
	cout << "------------------------" << endl;
	cout << "   List all compute nodes known in the current system" << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("session"))
    {
	cout << endl;
	cout << "list session" << endl;
	cout << "------------------------" << endl;
	cout << "   List all established database connection for distributed queries" << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("copy"))
    {
	cout << endl;
	cout << "list copy" << endl;
	cout << "------------------------" << endl;
	cout << "   List all executed table set copied" << endl;
	cout << endl;
    }
    else
    {
	cout << "Available help options:" << endl;
	cout << endl;
	cout << "   list tableset" << endl;
	cout << "   list datafile" << endl;
	cout << "   list archlog" << endl;
	cout << "   list user" << endl;
	cout << "   list lock" << endl;
	cout << "   list node" << endl;	
	cout << "   list session" << endl;
    }
}

void helpDefine()
{
    cout << endl;
    cout << "define tableset <tableset> <attr1=val1> <attr2=val2> ..." << endl;
    cout << "--------------------------------------------------------" << endl;
    cout << endl;
    cout << "   Define a new tableset. The following attributes must be specified:"<< endl;
    cout << endl;
    cout << "   tsroot          - Root directory of the database system and control files." << endl;
    cout << endl;
    cout << "   tsticket        - Name of the online backup tableset ticket." << endl;
    cout << endl;
    cout << "   syssize         - Size of the system space in number of pages." << endl;
    cout << endl;
    cout << "   tmpsize         - Size of the tempory space in number of pages." << endl;
    cout << endl;
    
    cout << "   logname         - Prefix of the configured redolog files." << endl;
    cout << endl;
    cout << "   logsize         - Size of a redolog file in bytes." << endl;
    cout << endl;
    cout << "   lognum          - Number of configured redolog files." << endl;
    cout << endl;
    cout << "   primary         - Name of the primary host." << endl;
    cout << endl;
    cout << "   secondary       - Name of the secondary host." << endl;
    cout << endl;
    cout << "   For a defintion statement the attributes syssize, tmpsize logsize and logname attributes are specified as integer values." << endl;
    cout << "   The attributes logfilename, meiator, primary and secondary  are specified as string values." << endl;
    cout << endl;
    cout << "   Sample:" << endl;
    cout << endl;
    cout << "      define tableset ts1 tsroot='/cego/ts1' syssize=100 tmpsize=100 logname='/cego/ts1/redolog'" << endl; 
    cout << "             logsize=100000 lognum=3 primary='geek' secondary='sam';" << endl;
    cout << endl;
}


void helpCreate()
{
    
    cout << endl;
    cout << "create tableset <tableset>" << endl;
    cout << "-------------------------" << endl;
    cout << endl;
    cout << "   Create the specified tableset as defined in the tableset description." << endl;
    cout << endl;
 
}

void helpStart()
{
    cout << endl;
    cout << "start tableset <tableset>" << endl;
    cout << "-------------------------" << endl;
    cout << endl;
    cout << "   Start the specified tableset on the configured primary node." << endl;
    cout << endl;
}

void helpCopy()
{
	
    cout << endl;
    cout << "copy tableset <tableset>" << endl;
    cout << "------------------------" << endl;
    cout << endl;
    cout << "   Copy the specified tableset from primary to secondary node." << endl;
    cout << endl;
}

void helpSwitch()
{
    cout << endl;
    cout << "switch tableset <tableset>" << endl;
    cout << "switch mediator for <tableset>" << endl;    
    cout << "switch secondary for <tableset>" << endl;
    cout << "------------------------------" << endl;
    cout << endl;
    cout << "   Switch the specified tableset to secondary node." << endl;
    cout << endl;
}

void helpStop()
{
    
    cout << endl;
    cout << "stop tableset <tableset>" << endl;
    cout << "------------------------" << endl;
    cout << endl;
    cout << "   Stop the specified tableset on the primary node." << endl;
    cout << endl;
}

void helpDrop()
{
    cout << endl;
    cout << "drop tableset <tableset>" << endl;
    cout << "------------------------" << endl;
    cout << endl;
    cout << "   Drop the specified tableset." << endl;
    cout << endl;
}

void helpRemove(const Chain& arg)
{

    if ( arg ==  Chain("tableset"))
    {
	cout << endl;
	cout << "remove tableset <tableset>" << endl;
	cout << "------------------------" << endl;
	cout << endl;
	cout << "   Remove the tableset defintion for the specified tableset." << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("archlog"))
    {
	cout << endl;
	cout << "remove archlog <archlog> from <tableset>" << endl;
	cout << "----------------------------------------" << endl;
	cout << "   Removes the specified archlog destination from the given tableset" << endl;
	cout << endl;
    }
    else if ( arg ==  Chain("role"))
    {
	cout << endl;
	cout << "remove role <role> from user <user> for <tableset>" << endl;
	cout << "--------------------------------------------------" << endl;
	cout << "   Removes the specified user role from tableset user" << endl;
	cout << endl;
    }
    else
    {
	cout << "Available help options:" << endl;
	cout << endl;
	cout << "   remove tableset" << endl;
	cout << "   remove archlog" << endl;
	cout << "   remove role" << endl;
	cout << endl;
    }
}

void helpSync(const Chain& arg)
{

    cout << endl;

    cout << "sync tableset <tableset> [ with '<command>' ]" << endl;
    cout << "------------------------------------" << endl;
    cout << endl;
    cout << "   Writes a checkpoint for the given tableset." << endl;
    cout << "   The optional given command is executed, during the tableset is synchronized and locked." << endl;
    cout << "   So this can be used to create a consistent snapshot for a tableset ( e.g. with ZFS snapshot commands )" << endl;
    cout << endl;
    
}

void helpImport(const Chain& arg)
{
    if ( arg == Chain("table"))
    {
	cout << endl;
	cout << "[ xml | binary ] import table <tablename> to <tableset> from '<filename>'" << endl;
	cout << "--------------------------------------------------------" << endl;
	cout << endl;
	cout << "   Import of a single table." << endl;
	cout << endl;
    }
    else if ( arg == Chain("tableset"))
    {
	cout << endl;
	cout << "[ xml | binary ] import tableset [ structure ] <tableset> from '<filename>'" << endl;
	cout << "------------------------------------------" << endl;
	cout << endl;
	cout << "   Import of a tableset to a file." << endl;
	cout << endl;
    }
    else
    {
	cout << "Available help options:" << endl;
	cout << endl;
	cout << "   import table" << endl;
	cout << "   import tableset" << endl;
    }   
}

void helpExport(const Chain& arg)
{

    if ( arg == Chain("table"))
    {
	cout << endl;
	cout << "[ xml | binary ] export table <tablename> from <tableset> to '<filename>'" << endl;
	cout << "--------------------------------------------------------" << endl;
	cout << endl;
	cout << "   Export of a single table." << endl;
	cout << endl;
    }
    else if ( arg == Chain("tableset"))
    {
	cout << endl;
	cout << "[ xml | binary ] export tableset [ structure ] <tableset> to '<filename>'" << endl;
	cout << "------------------------------------------" << endl;
	cout << endl;
	cout << "   Export of a tableset to a file." << endl;
	cout << endl;
    }
    else
    {
	cout << "Available help options:" << endl;
	cout << endl;
	cout << "   export table" << endl;
	cout << "   export tableset" << endl;
    }   
}

void helpAdd(const Chain& arg)
{
    if ( arg == Chain("user"))
    {
	cout << endl;
	cout << "add user <userid> identified by '<passwd>' to <tableset>" << endl;
	cout << "--------------------------------------------------------" << endl;
	cout << endl;
	cout << "   Starts online backup mode for the given tableset" << endl;
	cout << "   During backup mode, any datafile changes are detected" << endl;
	cout << "   If any recovery is required, the original page is restored during logfile recovery" << endl;
	cout << endl;
    }
    else if ( arg == Chain("datafile"))
    {
	cout << endl;
	cout << "add datafile '<filename>' size <numpages> to <tableset>" << endl;
	cout << "--------------------------------------------------------" << endl;
	cout << endl;
    }
    else if ( arg == Chain("archlog"))
    {
	cout << endl;
	cout << "add archlog <archid> path '<archpath>' to <tableset>" << endl;
	cout << "----------------------------------------------------" << endl;
	cout << endl;
    }
    else if ( arg == Chain("archlog"))
    {
	cout << endl;
	cout << "add role <role> to user <user> for <tableset>" << endl;
	cout << "---------------------------------------------" << endl;
	cout << "   Adds the specified role to tableset user" << endl;
	cout << endl;
	cout << "   Valid roles are" << endl;
	cout << endl;
	cout << "      ALL - All operations allowed" << endl;
	cout << "      READOBJ - Read operations are allowed" << endl;
	cout << "      WRITEOBJ - Write operations are allowed" << endl;
	cout << "      CREATEOBJ -  Object creation is allowed" << endl;
	cout << "      DROPOBJ - Object drop is allowed" << endl;
	cout << "      EXECPROC - Execution of strored procedures is allowed" << endl;
	cout << endl;
    }
    else
    {
	cout << "Available help options:" << endl;
	cout << endl;
	cout << "   add user" << endl;
	cout << "   add datafile" << endl;
	cout << "   add archlog" << endl;
	cout << "   add role" << endl;

    }   
}


void helpPasswd(const Chain& arg)
{
    cout << endl;
    cout << "passwd user <userid> identified by '<passwd>' to <tableset>" << endl;    
    cout << "-----------------------------------------------------------" << endl;
    cout << endl;
    cout << "   Set up the user passwd to the given value" << endl;
    cout << endl;
}

void helpAbort(const Chain& arg)
{

    cout << endl;
    cout << "abort db thread <tid>" << endl;
    cout << "---------------------" << endl;
    cout << endl;
    cout << "   Aborting the actual running query of an active db thread" << endl;
    cout << endl;
    
}

void helpBegin()
{
    cout << endl;
    cout << "begin backup for <tableset>" << endl;
    cout << "------------------------" << endl;
    cout << endl;
    cout << "   Starts online backup mode for the given tableset" << endl;
    cout << "   During backup mode, any datafile changes are detected" << endl;
    cout << "   If any recovery is required, the original page is restored during logfile recovery" << endl;
    cout << endl;
}

void helpEnd()
{
    cout << endl;
    cout << "end backup for <tableset>" << endl;
    cout << "------------------------" << endl;
    cout << endl;
    cout << "   Finishes online backup mode for the given tableset" << endl;
    cout << "   Normally this is done, after backup has been finished for all datafiles." << endl;
    cout << endl;
}

void helpRecover()
{
    cout << endl;
    cout << "recover tableset <tableset> [ to <point in time>]" << endl;
    cout << "-------------------------------------------------" << endl;
    cout << endl;
    cout << "   Execute an up-to-crash or point in time recover for the given tableset." << endl;
    cout << "   The corresponding tableset must be in status offline perform this action." << endl;
    cout << "   All required backup datafiles and offline redo logs have to be provided in the appropriate locations." << endl;
    cout << "   If an online backup should be restored, a valid backup ticket file has to be provided." << endl;
    cout << endl;
    cout << "   If point-in-time is used, an appropriate time stamp has to be provided in the following format" << endl;
    cout << endl;
    cout << "      <YYYY-mm-dd> HH:MM:SS Z>" << endl;
    cout << endl;
    cout << "   Sample : recover tableset TS1 to '2007-08-12 08:12:34 CEST';" << endl;
    cout << endl;
    cout << "   After the ceovery has been finished, the corresponding tableset can be started up." << endl;
    cout << endl;
}

void helpEnable()
{
    cout << endl;
    cout << "enable archlog for <tableset>" << endl;
    cout << "-----------------------------" << endl;
    cout << endl;
    cout << "   Enables archive log mode for the given tableset." << endl;
    cout << "   At least one valid archive log location must be defined to enable archive log mode" << endl;
    cout << endl;
}


void helpDisable()
{
    cout << endl;
    cout << "disable archlog for <tableset>" << endl;
    cout << "------------------------------" << endl;
    cout << endl;
    cout << "   Disables archive log mode for the given tableset." << endl;
    cout << endl;
}

void helpRelocate()
{
    cout << endl;
    cout << "relocate secondary for <tableset> to '<hostname>'" << endl;
    cout << "relocate mediator for <tableset> to '<hostname>'" << endl;
    cout << "---------------------------------------------" << endl;
    cout << endl;
    cout << "   Relocates the secondary or mediaotr role to the given hostname." << endl;
    cout << endl;
}

void helpCheck()
{
    cout << endl;
    cout << "check tableset <tableset>" << endl;
    cout << "---------------------------------------------" << endl;
    cout << endl;
    cout << "   Perform a consistency check for the given tableset" << endl;
    cout << endl;
}

