///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoClient.cc
// -------------
// Cego client program
//
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000-2010 Bjoern Lemke                        
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: main
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <lfc/Chain.h>
#include <lfc/Exception.h>
#include <lfc/GetLongOpt.h>
#include <lfc/File.h>
#include <lfc/Timer.h>
#include <lfc/Tokenizer.h>
#include <lfc/Net.h>
#include <lfc/SigHandler.h>
#include <lfc/Tokenizer.h>

#include <lfc/Element.h>
#include <lfc/Document.h>
#include <lfc/XMLSuite.h>

#include "CegoOutput.h"
#include "CegoViewObject.h"
#include "CegoProcObject.h"
#include "CegoKeyObject.h"
#include "CegoDbHandler.h"
#include "CegoModule.h"
#include "CegoNet.h"
#include "CegoXMLdef.h"

#include "CegoDefs.h"


#define USAGE "Usage: cgclt --user=<user>/<password>\n\
          [ --server=<host>]\n\
          [ --port=<port> ]\n\
	  [ --tableset=<tableset> ]\n\
          [ --batchfile=<batchfile> ]\n\
          [ --dumpfile=<dumpfile> ]\n\
          [ --cmd=<command> ]\n\
          [ --logfile=<logfile> ]\n\
          [ --raw ]\n\
          [ --ignore ]\n\
          [ --profile=<pid> ]\n\
          [ --debug ] [ --version  ] [ --help ]"


#define DEFAULTPORT 2200
#define DEFAULTSERVER "localhost"

#define HISTFILE ".cgclt_history"
#define PROMPT "CGCLT > "
#define PROMPT2 "> "

#define CGPROFILE ".cgprofile"

#define LANGUAGE_ENV "LANG"
#define LOCALE_CATEGORY LC_TIME

#define MAXCMDLEN 5000

enum RunMode { INTERACTIVE, BATCH, COMMAND, DUMP };

extern char __lfcVersionString[];
extern char __lfcxmlVersionString[];
extern char __quoteEscapeFlag;

bool flowCtl = false;
int flowCount = 10;

class AbortHandler : public SigHandler
{
  
public:

    AbortHandler(const Chain& serverName, 
		 int portNo,
		 const Chain& tableSet,
		 const Chain& user, 
		 const Chain& password,		 
		 long tid, 
		 CegoModule *pModule);
    ~AbortHandler();
    void sigCatch(int sig);

    void setQueryAbort(bool doAbort);
    
private:

    long _tid;
    bool _doAbort;
    CegoModule *_pModule;
    int _portNo;
    Chain _serverName;
    Chain _user;
    Chain _password;
    Chain _tableSet;
};

class ProfileEntry
{
  
public:

    ProfileEntry();
    ProfileEntry(const Chain& pid);
    ProfileEntry(const Chain& pid,
		 const Chain& serverName,
		 int portNo,
		 const Chain& tableSet,
		 const Chain& auth, 
		 const Chain& prompt);
    ~ProfileEntry();

    const Chain& getId() const;
    const Chain& getServerName() const;
    int getPortNo() const;
    const Chain& getTableSet() const;
    const Chain& getAuth() const;
    const Chain& getPrompt() const;
        
    ProfileEntry& operator = ( const ProfileEntry& ad);
    bool operator == ( const ProfileEntry& ad);
    
private:

    Chain _pid;
    int _portNo;
    Chain _serverName;
    Chain _tableSet;
    Chain _auth;
    Chain _prompt;
};

bool dispatchCmd(CegoDbHandler* pNetHandle, AbortHandler* pAH, Chain cmd, bool outMode, int & exitCode);
void loadProfile(const Chain& rcFileName, ListT<ProfileEntry>& pfeList);
void dumpTableSet(CegoNet *pCegoNet, const Chain& dumpFileName);
void dumpCounter(CegoNet *pCegoNet, File *pDF);
void dumpTable(CegoNet *pCegoNet, File *pDF);
void dumpIndex(CegoNet *pCegoNet, File *pDF);
void dumpKey(CegoNet *pCegoNet, File *pDF);
void dumpCheck(CegoNet *pCegoNet, File *pDF);
void dumpView(CegoNet *pCegoNet, File *pDF);
void dumpProcedure(CegoNet *pCegoNet, File *pDF);

int main(int argc, char **argv)
{	
    
    GetLongOpt longOpt(argc, argv);
    
    longOpt.addOpt("version");
    longOpt.addOpt("help");
    longOpt.addOpt("logfile");
    longOpt.addOpt("tableset");
    longOpt.addOpt("user");
    longOpt.addOpt("cmd");
    longOpt.addOpt("batchfile");
    longOpt.addOpt("dumpfile");
    longOpt.addOpt("qescmode");
    longOpt.addOpt("debug");
    longOpt.addOpt("raw");
    longOpt.addOpt("ignore");
    longOpt.addOpt("server", DEFAULTSERVER);
    longOpt.addOpt("port", Chain(DEFAULTPORT));
    longOpt.addOpt("profile");
    
    try
    {
	longOpt.parseOpt(); 
    }
    catch ( Exception e )
    {
	Chain msg;
	e.pop(msg);
	cerr << msg << endl;
	cerr << USAGE << endl;
	exit(1);	
    }
    
    Chain user;
    Chain password;
    

#ifdef HAVE_MINGW32
    Chain homeDir(getenv("HOMEPATH"));
    Chain histPath = homeDir + "\\" + Chain(HISTFILE);
#else
    Chain homeDir(getenv("HOME"));
    Chain histPath = homeDir + "/" + Chain(HISTFILE);
#endif
    
    RunMode runMode = INTERACTIVE;

    bool debug=false;
    if ( longOpt.isSet("debug") )
	debug=true;

    if ( longOpt.isSet("qescmode") )
	__quoteEscapeFlag=1;
    else
	__quoteEscapeFlag=0;

    bool rawMode = false;
    if ( longOpt.isSet("raw") )
	rawMode=true;

    bool ignoreError = false;
    if ( longOpt.isSet("ignore") )
	ignoreError=true;

    
    if ( longOpt.isSet( Chain("help") ) )
    {	
	cerr << USAGE << endl;
	exit(0);
    }

    if ( longOpt.isSet( Chain("version") ) )
    {
	cout << CEGO_PRODUCT << " (" << sizeof(long) * 8 << " bit), Version " << CEGO_VERSION 
	     << " [ lfc : " << __lfcVersionString  
	     << ", lfcxml : " <<  __lfcxmlVersionString << " ]" << endl;
	cout << CEGO_COPYRIGHT << endl;
	exit(0);
    }


    ListT<ProfileEntry> pfeList;

    loadProfile(homeDir + "/" + Chain(CGPROFILE), pfeList);

    Chain logFile = longOpt.getOptValue("logfile");

    Chain batchFileName;
    Chain dumpFileName;
    Chain command;

    if ( longOpt.isSet("batchfile") )
    {
	runMode = BATCH;
	batchFileName = longOpt.getOptValue("batchfile");
    }
    if ( longOpt.isSet("dumpfile") )
    {
	runMode = DUMP;
	dumpFileName = longOpt.getOptValue("dumpfile");
    }
    else if ( longOpt.isSet("cmd") )
    {
	runMode = COMMAND;
	command = longOpt.getOptValue("cmd");
    }


    Chain serverName;
    Chain tableSet;
    int portNo;
    Chain authString;
    Chain promptString;

    if ( longOpt.isSet("profile") )
    {
	Chain profile;
        profile = longOpt.getOptValue("profile");
	
	ProfileEntry *pE;
	if ( ( pE = pfeList.Find( ProfileEntry(profile))) != 0 )
	{
	    serverName = pE->getServerName();
	    tableSet = pE->getTableSet();
	    portNo = pE->getPortNo();
	    authString = pE->getAuth();
	    promptString = pE->getPrompt();	 
	}
	else
	{
	    cerr << "Profile entry " << profile << " not found" << endl;
	    exit(1);
	}
    }
    else
    {	
	serverName = longOpt.getOptValue("server");
	tableSet = longOpt.getOptValue("tableset");
	portNo = longOpt.getOptValue("port").asInteger();    
	authString = longOpt.getOptValue("user");
	promptString = Chain(PROMPT);	
    }

    Tokenizer authTok(authString, Chain("/")); 
    authTok.nextToken(user);
    authTok.nextToken(password);

    if ( user.length() == 0 )
    {
	cerr << "User not set" << endl;
	exit (1);
    }
    if ( password.length() == 0 )
    {
	cerr << "Password not set" << endl;
	exit (1);
    }
    if ( tableSet.length() == 0 )
    {
	cerr << "Tableset not set" << endl;
	exit (1);
    }

    // for localization
    char* lang = 0;
    if ( ( lang = getenv(LANGUAGE_ENV) ) != 0)
    {
	if ( setlocale(LOCALE_CATEGORY, lang) == 0)
	{
	    Chain msg = Chain("Cannot set locale ") + Chain(lang);
	    cerr << msg << endl;
	    exit(1);
	}
    }

    CegoModule *pModule;
    if ( logFile.length() > 0 )
	pModule = new CegoModule(logFile);
    else
	pModule = new CegoModule;

    
    unsigned long modId = pModule->getModId("CegoClient");
    
    if ( debug )
    {
	pModule->logModule(modId, "CegoClient", Logger::DEBUG);
	unsigned long id = pModule->getModId("CegoDbHandler");
	pModule->logModule(id, "CegoDbHandler", Logger::DEBUG);
    }
    else
    {
	pModule->logModule(modId, "CegoClient", Logger::NOTICE);
	unsigned long id = pModule->getModId("CegoDbHandler");
	pModule->logModule(id, "CegoDbHandler", Logger::NOTICE);
    }

    int exitCode;

    switch ( runMode )
    {

    case COMMAND:
    {
	
	Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	NetHandler* pN;
	
	pN = n.connect(serverName, portNo);
	
	CegoDbHandler* pSH = new CegoDbHandler(pN, pModule);
	
	CegoDbHandler::ResultType res = pSH->requestSession(tableSet, user, password);
	if ( res != CegoDbHandler::DB_OK )
	{
	    Chain msg;
	    pSH->getMsg(msg);
	    throw Exception(EXLOC, msg);
	}
			
	dispatchCmd(pSH, 0, command, rawMode, exitCode);

	pSH->closeSession();
	
	delete pSH;
	delete pN;
	
	exit(exitCode);
	
	break;
    }
    case BATCH:
    {
	
	int lineNo=0;
	    
	try 
	{

	    Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	    NetHandler* pN;
	    
	    pN = n.connect(serverName, portNo);

	    CegoDbHandler* pSH = new CegoDbHandler(pN, pModule);

	    CegoDbHandler::ResultType res = pSH->requestSession(tableSet, user, password);
	    if ( res != CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
		throw Exception(EXLOC, msg);
	    }
	    
	    File cmdFile(batchFileName);
	    
	    cmdFile.open(File::READ);
	    
	    Chain cmd;
	    Chain line;
	    bool disableDelimiter=false;
	    bool goOn = true;

	    while (goOn && cmdFile.readLine(line, MAXCMDLEN))
	    {
		lineNo++;
		line = line.cutTrailing(" \t");

		bool isCommentLine = false;
		
		if ( line.length() >= 2 )
		{		    
		    if (line.cutTrailing(" \t").subChain(1, 2) == Chain("--"))
		    {
			isCommentLine = true;
		    }
		}
		
		if ( ! isCommentLine )
		{

		    if ( line == Chain("@") )
		    {
			if ( disableDelimiter == false )
			    disableDelimiter=true;
			else
			    disableDelimiter=false;
		    }
		    else
		    {
			// cmd = line;
			cmd = cmd + Chain(" ") + line;
			cmd = cmd.cutTrailing(" \t");
		    }
			 
		    if ( cmd.length() > 0 )
		    {

			if (cmd.subChain(cmd.length()-1, cmd.length()) == Chain(";") 
			    && disableDelimiter==false)
			{
			    			
			    // send command to server here
				
			    goOn = dispatchCmd(pSH, 0, cmd, rawMode, exitCode);
			    
			    if ( exitCode != 0 && ignoreError )
				goOn = true;

			    cout << "ok" << endl;			    			    			    
			    
			    cmd = Chain("");
			    
			}
		    }
		}
	    }
	    if ( cmd.length() > 1 )
	    {
		cout << "incomplete command ..<" << cmd << ">" << endl;
	    }
	    
	    cmdFile.close();

	    pSH->closeSession();

	    delete pSH;
	    delete pN;

	}
	catch ( Exception e)
	{
	    cout << "Line Number " + Chain(lineNo) + " : " << endl;
	    e.print();
	    exitCode=1;
	}
	break;
    }
    case DUMP:
    {
	
	Chain logMode("notice");
	if ( debug )
	    logMode=Chain("debug");

	CegoNet *pCegoNet = new CegoNet( logFile, logMode );	
	pCegoNet->connect(serverName, portNo, tableSet, user, password);
	
	dumpTableSet(pCegoNet, dumpFileName);
	
	delete pCegoNet;
	break;
    }
    case INTERACTIVE:
    {
	try 
	{	    	    


	    Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	    NetHandler* pN;

	    pN = n.connect(serverName, portNo);

	    CegoDbHandler* pSH = new CegoDbHandler(pN, pModule);

	    pModule->log(modId, Logger::DEBUG, Chain("Requesting session ..."));

	    CegoDbHandler::ResultType res = pSH->requestSession(tableSet, user, password);
	    if ( res != CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
		throw Exception(EXLOC, msg);
	    }

	    long tid = pSH->getTid();

	    AbortHandler *pAH = new AbortHandler(serverName, portNo, tableSet, user, password, tid, pModule);
	    
	    pModule->log(modId, Logger::DEBUG, Chain("Got session ..."));
	    read_history((char*)histPath);
 
	    bool goOn = true;
	    while ( goOn )
	    {
		
		char *pC = 0;
		
		Chain cs;
		
		pC = readline(promptString);
		
		if (pC)
		{
		    cs = Chain(pC);

		    bool stmtComplete = false;
		    while (! stmtComplete )
		    {		
			add_history(pC);	
			
			if (pC)
			    free(pC);	    
			
			cs = cs.cutTrailing(Chain(" \t"));
			
			if ( cs.length() > 1)
			{		
			    if ( cs[cs.length()-2] == ';')
				stmtComplete = true;
			}
			
			if ( ! stmtComplete )
			{
			    if ( cs.length() > 1 )
			    {
				pC = readline(PROMPT2);
			    }
			    else
			    {
				pC = readline(promptString);
			    }
			    if (pC)
				cs = cs + Chain(" ") + Chain(pC);	   
			}
			
		    }
		    
		    Timer t(6,3);
		    t.start();

		    goOn = dispatchCmd(pSH, pAH, cs, rawMode, exitCode);

		    if ( exitCode != 0 )
			goOn = true;
				
		    // sending command to server here
		    
		    t.stop();
		    cerr << "ok ( " << t << " s )" << endl;
		    
		    write_history((char*)histPath);
		}
		else
		{
		    cout << "Bye" << endl;
		    goOn=false;
		}
	    }
	    
	    pSH->closeSession();

	    delete pSH;
	    delete pN;
	    delete pAH;
	    
	}
	catch ( Exception e)
	{
	    Chain msg;
	    e.pop(msg);
	    cerr << msg << endl;
	    exitCode=1;	   
	}
	break;
    }    
    }

    delete pModule;
    exit(exitCode);
}

bool dispatchCmd(CegoDbHandler *pSH, AbortHandler *pAH, Chain cmd, bool outMode, int& exitCode)
{
    
    exitCode = 0;

    if ( cmd.cutTrailing(" ;") == Chain("quit") )
	return false;

    Tokenizer tok(cmd, Chain(" ;"));
    Chain flowTok;
    if ( tok.nextToken(flowTok) )
    {
	
	if ( flowTok == Chain("flowctl") )
	{
	    Chain flowOpt;
	    if ( tok.nextToken(flowOpt) )
	    {
		if ( flowOpt == Chain("off") )
		{
		    flowCtl=false;
		    cout << "Flow control disabled" << endl;
		}
		else
		{
		    flowCtl=true;
		    if ( flowOpt.asInteger() > 0 )
			flowCount=flowOpt.asInteger();
		    cout << "Flow control enabled" << endl;    
		}
	    }	    
	    return true;
	}
    }
    
    Chain msg;
    CegoDbHandler::ResultType res;

    if ( pAH )
	pAH->setQueryAbort(true);
    
    res = pSH->reqQueryOp(cmd);
    
    if ( pAH )
	pAH->setQueryAbort(false);


    if ( res == CegoDbHandler::DB_OK )
    {
	Chain msg;
	pSH->getMsg(msg);
	long affCount = pSH->getAffected();

	// if ( msg == Chain("Goodbye") )
	//     return true;

	cout << msg << endl; // " ( " << affCount << " affected )" << endl;
	
    }
    else if ( res == CegoDbHandler::DB_INFO )
    {

	CegoObject::ObjectType objType = pSH->getObjType();

	switch ( objType )
	{
	case CegoObject::SYSTEM:
	case CegoObject::TABLE:
	case CegoObject::PINDEX:
	case CegoObject::UINDEX:
	case CegoObject::INDEX:
	{   
	    CegoTableObject to;
	    to.putElement( pSH->getObjElement() );
	    cout << to.getFormatted();
	    break;
	}
	case CegoObject::VIEW:
	{
	    CegoViewObject vo;
	    vo.putElement( pSH->getObjElement() );
	    cout << vo.getFormatted();
	    break;
	}
	case CegoObject::PROCEDURE:
	{
	    CegoProcObject po;
	    po.putElement( pSH->getObjElement() );
	    cout << po.getFormatted();
	    break;
	}
	case CegoObject::FKEY:
	{
	    CegoKeyObject ko;
	    ko.putElement( pSH->getObjElement() );
	    cout << ko.getFormatted();
	    break;
	}
	default:
	    break;
	}
    }
    else if ( res == CegoDbHandler::DB_DATA )
    {
	ListT<CegoField> schema;
	Chain format;

	pSH->getSchema(schema);
	pSH->getFormat(format);

	CegoOutput output(schema, format);
	
	output.setRawMode(outMode);

	output.headOut();
	
	// CegoDbHandler::QuerySync querySync;

	// querySync=CegoDbHandler::QUERY_MOREDATA;

	ListT<CegoField> fvl;
	long count=0;
	
	while ( ( res = pSH->receiveTableData(schema, fvl) ) == CegoDbHandler::DB_DATA )
	{	    
	    count++;
	    
	    output.rowOut(fvl);
	    fvl.Empty();		    

	    if ( flowCtl )
	    {
		if ( count % flowCount == 0 )
		{
		    cout << "Next, Reset, Abort ? ( n, r, a ) ";
		    char resp;
		    cin >> resp;
		    if ( resp == 'a' )
		    {
			pSH->abortQuery();
			break;
		    }
		    if ( resp == 'r' )
		    {
			pSH->resetQuery();
		    }
		    else
		    {
			// nothing
		    }		    
		}
	    }
		
	}
	output.tailOut();

    }
    if ( res == CegoDbHandler::DB_ERROR )
    {
	Chain msg;
	pSH->getMsg(msg);
	cout << "Error : " << msg << endl;
	cout << "Query failed" << endl;
	exitCode = 1;
	return false;
	
    }

    return true;
}

void dumpTableSet(CegoNet *pCegoNet, const Chain& dumpFileName)
{
 
    File* pDF = new File(dumpFileName);
    pDF->open(File::WRITE);

    try
    {
	dumpCounter(pCegoNet, pDF);	
	dumpTable(pCegoNet, pDF);
	dumpIndex(pCegoNet, pDF);
	dumpKey(pCegoNet, pDF);
	dumpCheck(pCegoNet, pDF);
	dumpView(pCegoNet, pDF);
	dumpProcedure(pCegoNet, pDF);
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);
	cerr << msg << endl;
    }

    pDF->close();
    delete pDF;

}

void dumpCounter(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list counter;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain counterName = pF->getValue().valAsChain();
	pF = fvl.Next();
	Chain counterValue = pF->getValue().valAsChain();

	Chain cmd;

	cmd=Chain("drop if exists counter ") + counterName + Chain(";\n");
	pDF->writeChain(cmd);


	cmd=Chain("create counter ") + counterName + Chain(";\n");
	pDF->writeChain(cmd);
	cmd=Chain("set counter ") + counterName + Chain(" to ") + counterValue + Chain(";\n");
	pDF->writeChain(cmd);
	fvl.Empty();	
    }
}

void dumpTable(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list table;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<Chain> tableList;
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain tableName = pF->getValue().valAsChain();
	tableList.Insert(tableName);
	fvl.Empty();	
    }
    
    Chain *pTable = tableList.First();
    while ( pTable ) 
    {
	res = pCegoNet->doQuery(Chain("desc table ") + *pTable + Chain(";"));

	Chain cmd;

	cmd=Chain("drop if exists table ") + *pTable + Chain(";\n");
	pDF->writeChain(cmd);
	
	cmd=Chain("create table ") + *pTable + Chain("(");

	ListT<CegoField> schema;
	pCegoNet->getSchema(schema);
    
	ListT<CegoField> fvl;
	bool isFirst=true;
	while ( pCegoNet->fetchData(schema, fvl)  )
	{
	    if ( isFirst == false )
		cmd += Chain(", ");
	    isFirst=false;

	    CegoField *pF = fvl.First();
	    Chain attrName = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    Chain attrType = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    Chain attrSize = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    Chain attrDef = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    Chain attrNull = pF->getValue().valAsChain();

	    cmd += attrName + Chain(" ") + attrType;
	    if ( attrType == Chain(XML_STRING_VALUE)
		 || attrType == Chain(XML_BIGINT_VALUE)
		 || attrType == Chain(XML_DECIMAL_VALUE))
		cmd +=  Chain("(") + attrSize + Chain(")");
	    if ( attrDef != Chain(XML_NULL_VALUE))
		cmd += Chain(" default ") + attrDef;
	    if ( attrNull == Chain("no"))
		cmd += Chain(" not null");
	    fvl.Empty();	
	}

	cmd += Chain(");\n");
	pDF->writeChain(cmd);


	// dump table content
	res = pCegoNet->doQuery(Chain("select * from ") + *pTable + Chain(";"));

	schema.Empty();
	fvl.Empty();

	pCegoNet->getSchema(schema);
    
	while ( pCegoNet->fetchData(schema, fvl)  )
	{

	    Chain cmd=Chain("insert into ") + *pTable + Chain(" values (");

	    CegoField *pF = fvl.First();
	    while ( pF )
	    {
		Chain val = pF->getValue().toChain();
		cmd += val;
		pF = fvl.Next();
		
		if ( pF )
		    cmd += Chain(", ");
	    }
	    cmd += Chain(");\n");
	    pDF->writeChain(cmd);
	    fvl.Empty();
	}

	pTable = tableList.Next();
    }
}


void dumpIndex(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list index;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<Chain> indexList;
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain indexName = pF->getValue().valAsChain();
	indexList.Insert(indexName);
	fvl.Empty();	
    }
    
    Chain *pIndex = indexList.First();
    while ( pIndex ) 
    {
	res = pCegoNet->doQuery(Chain("desc index ") + *pIndex + Chain(";"));


	ListT<CegoField> schema;
	pCegoNet->getSchema(schema);
    
	Chain tableName;
	Chain indexType;
	ListT<Chain> attrList;
	
	ListT<CegoField> fvl;	
	while ( pCegoNet->fetchData(schema, fvl)  )
	{	    
	    CegoField *pF = fvl.First();
	    Chain attrName = pF->getValue().valAsChain();
	    attrList.Insert(attrName);
	    pF = fvl.Next();
	    tableName = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    indexType = pF->getValue().valAsChain();

	    fvl.Empty();
	}

	Chain cmd;

	cmd=Chain("drop if exists index ") + *pIndex + Chain(";\n");
	pDF->writeChain(cmd);

	if ( indexType == Chain(XML_PINDEX_VALUE) )
	    cmd=Chain("create primary index ");
	else if ( indexType == Chain(XML_UINDEX_VALUE) )
	    cmd=Chain("create unique index ") + *pIndex; 
	else
	    cmd=Chain("create index ") + *pIndex;

	cmd += Chain(" on ") + tableName + Chain("(");
	
	Chain *pAttr = attrList.First();
	bool isFirst=true;
	while ( pAttr )
	{
	    if ( isFirst == false )
		cmd += Chain(",");
	    isFirst=false;

	    cmd += *pAttr;
	    pAttr = attrList.Next();
	}
	
	cmd += Chain(");\n");
	pDF->writeChain(cmd);

	pIndex = indexList.Next();
    }
}


void dumpKey(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list key;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<Chain> keyList;
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain keyName = pF->getValue().valAsChain();
	keyList.Insert(keyName);
	fvl.Empty();	
    }
    
    Chain *pKey = keyList.First();
    while ( pKey ) 
    {
	res = pCegoNet->doQuery(Chain("desc key ") + *pKey + Chain(";"));
	
	ListT<CegoField> schema;
	pCegoNet->getSchema(schema);

	Chain keyTable;
	Chain refTable;
	ListT<Chain> keyAttrList;
	ListT<Chain> refAttrList;

	ListT<CegoField> fvl;
	while ( pCegoNet->fetchData(schema, fvl)  )
	{

	    CegoField *pF = fvl.First();

	    Chain tableName = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    Chain attrName = pF->getValue().valAsChain();
	    pF = fvl.Next();
	    Chain attrType = pF->getValue().valAsChain();
	    
	    if ( attrType == Chain("key") )
	    {
		keyTable=tableName;
		keyAttrList.Insert(attrName);
	    }
	    else if ( attrType == Chain("reference") )
	    {
		refTable=tableName;
		refAttrList.Insert(attrName);		
	    }

	    fvl.Empty();	
	}

	Chain cmd;	

	cmd=Chain("drop if exists foreign key ") + *pKey + Chain(";\n");
	pDF->writeChain(cmd);


	cmd=Chain("alter table ") + keyTable + Chain(" add foreign key ") + *pKey + Chain("(");
	Chain *pKeyAttr = keyAttrList.First();
	while ( pKeyAttr )
	{
	    cmd += *pKeyAttr;
	    pKeyAttr = keyAttrList.Next();
	    if ( pKeyAttr )
		cmd += Chain(",");
	}

	cmd+=Chain(") references ") + refTable + Chain("(");
	Chain *pRefAttr = refAttrList.First();
	while ( pRefAttr )
	{
	    cmd += *pRefAttr;
	    pRefAttr = refAttrList.Next();
	    if ( pRefAttr )
		cmd += Chain(",");
	}

	cmd += Chain(");\n");
	pDF->writeChain(cmd);

	pKey = keyList.Next();
    }
}

void dumpCheck(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list check;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<Chain> checkList;
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain checkName = pF->getValue().valAsChain();
	checkList.Insert(checkName);
	fvl.Empty();	
    }
    
    Chain *pCheck = checkList.First();
    while ( pCheck ) 
    {
	res = pCegoNet->doQuery(Chain("desc check ") + *pCheck + Chain(";"));


	ListT<CegoField> schema;
	pCegoNet->getSchema(schema);
    
	ListT<CegoField> fvl;
	bool isFirst=true;
	while ( pCegoNet->fetchData(schema, fvl) );

	CegoField *pF = fvl.First();
	Chain checkTable = pF->getValue().valAsChain();
	pF = fvl.Next();
	Chain checkCondition = pF->getValue().valAsChain();
	
	fvl.Empty();	
       
	Chain cmd;	

	cmd=Chain("drop if exists check ") + *pCheck + Chain(";\n");
	pDF->writeChain(cmd);

	cmd=Chain("alter table ") +  checkTable + Chain(" add check ") + *pCheck + Chain(" on ") + checkCondition + Chain(";\n");
	pDF->writeChain(cmd);

	pCheck = checkList.Next();
    }
}

void dumpView(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list view;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<Chain> viewList;
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain viewName = pF->getValue().valAsChain();
	viewList.Insert(viewName);
	fvl.Empty();	
    }
    
    Chain *pView = viewList.First();
    while ( pView ) 
    {
	res = pCegoNet->doQuery(Chain("show view ") + *pView + Chain(";"));


	ListT<CegoField> schema;
	pCegoNet->getSchema(schema);
    
	ListT<CegoField> fvl;
	bool isFirst=true;
	Chain viewText;
	while ( pCegoNet->fetchData(schema, fvl)  )
	{
	    CegoField *pF = fvl.First();
	    viewText = pF->getValue().valAsChain();
	    fvl.Empty();	
	}

	Chain cmd;	

	cmd=Chain("drop if exists view ") + *pView + Chain(";\n");
	pDF->writeChain(cmd);

	cmd=Chain("create ") + viewText;
	cmd += Chain("\n");
	pDF->writeChain(cmd);

	pView = viewList.Next();
    }
}

void dumpProcedure(CegoNet *pCegoNet, File *pDF)
{
	
    long res;
    res = pCegoNet->doQuery(Chain("list procedure;"));
        
    ListT<CegoField> schema;
    pCegoNet->getSchema(schema);
    
    ListT<Chain> procList;
    ListT<CegoField> fvl;
    while ( pCegoNet->fetchData(schema, fvl)  )
    {
	CegoField *pF = fvl.First();
	Chain procName = pF->getValue().valAsChain();
	procList.Insert(procName);
	fvl.Empty();	
    }
    
    Chain *pProc = procList.First();
    while ( pProc ) 
    {
	res = pCegoNet->doQuery(Chain("show procedure ") + *pProc + Chain(";"));


	ListT<CegoField> schema;
	pCegoNet->getSchema(schema);
    
	ListT<CegoField> fvl;
	Chain procText;
	while ( pCegoNet->fetchData(schema, fvl)  )
	{
	    CegoField *pF = fvl.First();
	    procText = pF->getValue().valAsChain();
	    fvl.Empty();	
	}

	Chain cmd;

	cmd=Chain("drop if exists procedure ") + *pProc + Chain(";\n");
	pDF->writeChain(cmd);

	cmd=Chain("@\ncreate ") + procText;
	cmd += Chain("\n@\n");
	pDF->writeChain(cmd);

	pProc = procList.Next();
    }
}

void loadProfile(const Chain& rcFileName, ListT<ProfileEntry>& pfeList)
{
    File rcFile(rcFileName);
    if ( rcFile.exists() )
    {
	
	rcFile.open(File::READ);

	Chain rcLine;
	while ( rcFile.readLine(rcLine) )
	{
	    Tokenizer rc1(rcLine, Chain(":"));
	    Chain pid;
	    Chain rcEntry;
	    rc1.nextToken(pid);
	    rc1.nextToken(rcEntry);
	    
	    Tokenizer rc2(rcEntry, Chain(","));
	    
	    Chain serverName;
	    int portNo;
	    Chain tableSet;
	    Chain auth;
	    Chain prompt;
	    
	    Chain rcField;
	    while ( rc2.nextToken(rcField) )
	    {
		Tokenizer rc3(rcField, Chain("="));
		Chain attr;
		Chain val;
		rc3.nextToken(attr);
		rc3.nextToken(val);
		if ( attr == Chain("server") )
		    serverName = val;
		else if ( attr == Chain("port") )
		    portNo = val.asInteger();
		else if ( attr == Chain("tableset") )
		    tableSet = val;
		else if ( attr == Chain("auth") )
		    auth = val;
		else if ( attr == Chain("prompt") )
		    prompt = val;
	    }

	    ProfileEntry pfe(pid, serverName, portNo, tableSet, auth, prompt);
	    pfeList.Insert(pfe);
	    
	}		
    }
}

AbortHandler::AbortHandler(const Chain& serverName, 
			   int portNo, 
			   const Chain& tableSet,
			   const Chain& user, 
			   const Chain& password,
			   long tid, 
			   CegoModule *pModule) :  SigHandler() 
{
    init();
    install(SIGINT);
#ifndef HAVE_MINGW32
    install(SIGPIPE);
#endif
    _tid = tid;
    _pModule = pModule;
    _portNo = portNo;
    _tableSet = tableSet;
    _user = user;
    _password = password;
    _serverName = serverName;
}

AbortHandler::~AbortHandler()
{
} 

void AbortHandler::sigCatch(int sig) 
{
    
    try {
	install(SIGINT);
#ifndef HAVE_MINGW32
	install(SIGPIPE);
#endif

	if ( _doAbort )
	{

	    cout << "Aborting query .." << endl;

	    Net n (  NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	    NetHandler* pN;
	    
	    pN = n.connect(_serverName, _portNo);
	    
	    CegoDbHandler* pSH = new CegoDbHandler(pN, _pModule);
	    	    
	    CegoDbHandler::ResultType res = pSH->requestSession(_tableSet, _user, _password);
	    if ( res != CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
		throw Exception(EXLOC, msg);
	    }
    
	    Chain msg;
	    	    
	    res = pSH->reqQueryAbort(_tid);
	    	    
	    if ( res == CegoDbHandler::DB_OK )
	    {
		Chain msg;
		pSH->getMsg(msg);
	    }

	    
	    pSH->closeSession();

	    delete pSH;
	    delete pN;

	}
	else
	{
	    cout << "Use quit to quit" << endl;
	}
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);
	cout << "Error : " << msg << endl;
    }
    
}

void AbortHandler::setQueryAbort(bool doAbort)
{
    _doAbort = doAbort;
}


ProfileEntry::ProfileEntry()
{
}

ProfileEntry::ProfileEntry(const Chain& pid)
{
    _pid = pid;
}

ProfileEntry::ProfileEntry(const Chain& pid, 
			   const Chain& serverName, 
			   int portNo, 
			   const Chain& tableSet,
			   const Chain& auth, 
			   const Chain& prompt)
{
    _pid = pid;
    _serverName = serverName;
    _portNo = portNo;
    _tableSet = tableSet;
    _auth = auth;
    _prompt = prompt;
}

ProfileEntry::~ProfileEntry()
{
} 


const Chain& ProfileEntry::getId() const
{
    return _pid;
}

const Chain& ProfileEntry::getServerName() const
{
    return _serverName;
}

int ProfileEntry::getPortNo() const
{
    return _portNo;
}

const Chain& ProfileEntry::getTableSet() const
{
    return _tableSet;
}

const Chain& ProfileEntry::getAuth() const
{
    return _auth;
}

const Chain& ProfileEntry::getPrompt() const
{
    return _prompt;
}

ProfileEntry& ProfileEntry::operator = ( const ProfileEntry& pe)
{
    _pid = pe._pid;
    _serverName = pe._serverName;
    _portNo = pe._portNo;
    _tableSet = pe._tableSet;
    _auth = pe._auth;
    _prompt = pe._prompt;
    return (*this);
}
    
bool ProfileEntry::operator == ( const ProfileEntry& pe)
{
    if ( _pid == pe._pid ) 
	return true;
    return false;
}
