///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoDbHandler.cc
// -----------------
// Cego db handler class implementation
//                                                         
// Design and Implementation by Bjoern Lemke               
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoDbHandler
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

#include <lfc/AESCrypt.h>
#include <lfc/Datetime.h>

#include <lfc/Element.h>

#include "CegoTypeConverter.h"
#include "CegoDbHandler.h"
#include "CegoXMLdef.h"
#include "CegoXMLHelper.h"
#include "CegoDefs.h"

#include <string.h>
#include <stdlib.h>

#define QUERY_MOREDATA 0
#define QUERY_ABORT 1
#define QUERY_RESET 2

CegoDbHandler::CegoDbHandler(NetHandler *pN, CegoModule *pModule)
{
    _pN = pN;
    _pModule = pModule;
    _pRow = 0;

    Document *pDoc = new Document(XML_CEGO);
    pDoc->setAttribute(XML_VERSION_ATTR, XML_VERSION_VALUE);
    
    _xml.setDocument(pDoc);

    _wasReset = false;

    _modId = _pModule->getModId("CegoDbHandler");
}

CegoDbHandler::~CegoDbHandler()
{
    Document *pDoc = _xml.getDocument();
    pDoc->clear();
    delete pDoc;
}

CegoDbHandler::ResultType CegoDbHandler::requestSession(const Chain& tableSet, const Chain& user, const Chain& password, bool doEncrypt)
{   

    _pModule->log(_modId, Logger::DEBUG, Chain("Request session for user ") + user + Chain("/") + password);
    Element* pRoot = new Element(XML_FRAME_ELEMENT);

    pRoot->setAttribute(XML_TABLESET_ATTR, tableSet);   
    pRoot->setAttribute(XML_USER_ATTR, user);

    _tableSet = tableSet;

    if ( doEncrypt )
    { 
	AESCrypt aescrypt(CEGOAESKEY, CEGOAESKEYLEN);
	pRoot->setAttribute(XML_PASSWD_ATTR, aescrypt.encrypt(password));
    }
    else
    {
	pRoot->setAttribute(XML_PASSWD_ATTR, password);
    }
    return sendReq(XML_DBSESSION_REQUEST, pRoot);
}

CegoDbHandler::ResultType CegoDbHandler::closeSession()
{        
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    return sendReq(XML_SESSION_CLOSE, pRoot);
}

bool CegoDbHandler::acceptSession()
{	    
    _pModule->log(_modId, Logger::DEBUG, Chain("Accepting session"));
            
    try {

	_xml.getDocument()->clear();
	_xml.setChain( _pN->getMsg() );	
	_pModule->log(_modId, Logger::DEBUG, _pN->getMsg());

	_xml.parse();

	_pModule->log(_modId, Logger::DEBUG, Chain("XML parse ok"));
	
	Chain docType = _xml.getDocument()->getDocType();
	
	if ( docType != Chain(XML_DBSESSION_REQUEST) )
	{
	    
	    _pModule->log(_modId, Logger::LOGERR, Chain("Invalid request"));

	    _xml.getDocument()->clear();
	    
	    Element* pRoot = new Element(XML_FRAME_ELEMENT);
	    pRoot->setAttribute(XML_MSG_ATTR, Chain("Invalid request"));

	    _xml.getDocument()->setRootElement(pRoot);
	    _xml.getDocument()->setDocType(XML_ERROR_DOC);
	    	    
	    Chain request;
	    _xml.getXMLChain(request);
	    
	    _pN->setMsg(request, request.length());
	    
	    _pN->writeMsg();
    
	    return false;
	}
	else
	{
	    Element *pRoot = _xml.getDocument()->getRootElement();
	    if ( pRoot )
	    {
		_tableSet = pRoot->getAttributeValue(XML_TABLESET_ATTR);
		_user = pRoot->getAttributeValue(XML_USER_ATTR);
		_password = pRoot->getAttributeValue(XML_PASSWD_ATTR);
		
		// response is sent after authorization check

		return true;
	    }
	    else
	    {
		throw Exception(EXLOC, "Cannot get root element from message");
	    }	    
	}
    }
    catch ( Exception e)
    {
		
	Chain msg;
	e.pop(msg);
	
	_pModule->log(_modId, Logger::LOGERR, Chain("Aborting session. Reason=") + msg);
	
	_pN->sendNack();
	return false;
	
    }
}

CegoDbHandler::ResultType CegoDbHandler::reqQueryOp(const Chain& cmd)
{

    _xml.getDocument()->clear();
    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_CMD_ATTR, cmd);

    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(XML_QUERY_REQUEST);
    
    Chain request;
    _xml.getXMLChain(request);
    
    _pN->setMsg(request, request.length());
    
    _pN->writeMsg();
    
    _pN->readMsg();

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();
    
    if ( docType == Chain(XML_OK_DOC) )
    {
	return DB_OK;
    }
    else if ( docType == Chain(XML_ERROR_DOC) )
    {
	return DB_ERROR;
    }
    else if ( docType == Chain(XML_DATA_DOC) )
    {
	return DB_DATA;
    }
    else if ( docType == Chain(XML_INFO_DOC) )
    {
	return DB_INFO;
    }
    else
    {
	throw Exception(EXLOC, "Invalid document type");
    }
}

CegoDbHandler::ResultType CegoDbHandler::reqQueryAbort(long idx)
{    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_TID_ATTR, Chain(idx));
    return sendReq(XML_QUERYABORT_REQUEST, pRoot);
}

void CegoDbHandler::getMsg(Chain& msg)
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	msg = pRoot->getAttributeValue(XML_MSG_ATTR);
    }
}


long CegoDbHandler::getAffected()
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	return pRoot->getAttributeValue(XML_AFFECTED_ATTR).asLong();
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}

long CegoDbHandler::getTid()
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	return pRoot->getAttributeValue(XML_TID_ATTR).asLong();
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}

int CegoDbHandler::getTabSetId()
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	return pRoot->getAttributeValue(XML_TSID_ATTR).asInteger();
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}

const Chain& CegoDbHandler::getTableSet()
{
    return _tableSet;
}

const Chain& CegoDbHandler::getUser()
{
    return _user;
}

const Chain& CegoDbHandler::getPassword()
{
    return _password;
}

void CegoDbHandler::getTableName(Chain& tableName)
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	tableName = pRoot->getAttributeValue(XML_TABLENAME_ATTR);
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}

void CegoDbHandler::getObjName(Chain& objName)
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	objName = pRoot->getAttributeValue(XML_OBJNAME_ATTR);
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}


void CegoDbHandler::getFormat(Chain& format)
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	format = pRoot->getAttributeValue(XML_FORMAT_ATTR);
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}

void CegoDbHandler::getSchema(ListT<CegoField>& schema)
{

    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	
	Chain tableName = pRoot->getAttributeValue(XML_TABLENAME_ATTR);
	
	ListT<Element*> childList = pRoot->getChildren(Chain(XML_SCHEMA_ELEMENT));
	Element **pRow = childList.First();
	while ( pRow ) 
	{
	    
	    Chain colTable = (*pRow)->getAttributeValue(XML_TABLENAME_ATTR);
	    Chain colName = (*pRow)->getAttributeValue(XML_COLNAME_ATTR);
	    Chain colType = (*pRow)->getAttributeValue(XML_COLTYPE_ATTR);
	    Chain colSize = (*pRow)->getAttributeValue(XML_COLSIZE_ATTR);
	    Chain colNullable = (*pRow)->getAttributeValue(XML_COLNULLABLE_ATTR);
	    Chain colDefValue = (*pRow)->getAttributeValue(XML_COLDEFVALUE_ATTR);
	    
	    CegoTypeConverter tc;
	    CegoDataType dataType = tc.getTypeId(colType);
	    
	    bool isNullable;
	    if ( colNullable == Chain(XML_TRUE_VALUE) )
		isNullable=true;
	    else
		isNullable=false;
		 
	    CegoFieldValue defValue;
	    if ( colDefValue != Chain("") )
	    {
		defValue = CegoFieldValue(dataType, colDefValue);
	    }
	    CegoField f(CegoField(colTable, colTable, colName, dataType, colSize.asInteger(), defValue, isNullable));	
	    
	    schema.Insert(f);
	    
	    pRow = childList.Next();
	}
    }
    else
    {
	throw Exception(EXLOC, "No root element found");
    }
}

Element* CegoDbHandler::getObjElement()
{
    Element *pRoot = _xml.getDocument()->getRootElement();
    
    ListT<Element*> objElementList = pRoot->getChildren(XML_OBJ_ELEMENT);

    Element **pOE = objElementList.First();
    if ( pOE ) 
    {
	return *pOE;
    }
    
    return 0;
}

CegoObject::ObjectType CegoDbHandler::getObjType()
{

    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	ListT<Element*> objElementList = pRoot->getChildren(XML_OBJ_ELEMENT);
	
	Element **pOE = objElementList.First();
	if ( pOE ) 
	{
	    CegoTypeConverter tc;
	    return tc.getObjectTypeId( (*pOE)->getAttributeValue(XML_OBJTYPE_ATTR) );
	}
	throw Exception(EXLOC, "No object type id found");	
    }
    throw Exception(EXLOC, "No root element found");	

}

void CegoDbHandler::getProcResult(ListT<CegoProcVar>& outParamList, CegoFieldValue& retValue)
{

   Element *pRoot = _xml.getDocument()->getRootElement();
   
   if ( pRoot )
   {
       ListT<Element*> outParamElementList = pRoot->getChildren(XML_OUTPARAM_ELEMENT);
       
       ListT<CegoField> fl;
       Element **pOP = outParamElementList.First();
       while ( pOP ) 
       {
	   Chain paramName = (*pOP)->getAttributeValue(XML_NAME_ATTR);
	   Chain paramValue = (*pOP)->getAttributeValue(XML_VALUE_ATTR);
	   Chain paramType = (*pOP)->getAttributeValue(XML_TYPE_ATTR);
	   	  
	   CegoTypeConverter tc;
	   CegoDataType datatype = tc.getTypeId(paramType);
	    
	   if ( paramName == Chain("") )
	   {
	       retValue = CegoFieldValue(datatype, paramValue);
	   }
	   else
	   {
	       CegoFieldValue fv(datatype, paramValue);
	       outParamList.Insert( CegoProcVar(paramName, CegoProcVar::OUTVAR, fv.getType(), fv.getLength(), fv));
	   }
	   
	   pOP = outParamElementList.Next();
       }
   }
   else
   {
       throw Exception(EXLOC, "No root element found");
   }
}

CegoDbHandler::ResultType CegoDbHandler::receiveTableData(const ListT<CegoField>& schema, ListT<CegoField>& fvl)
{

    _pModule->log(_modId, Logger::DEBUG, Chain("Receiving table data ..."));

    if ( _pRow )
	_pRow = _rowList.Next();

    if ( _pRow == 0 )
    {
	// get more data

	ResultType res = getMoreTableData();

	switch ( res )
	{
	case DB_DATA:
	    _pRow = _rowList.First();
	    break;
	case DB_OK:
	case DB_ERROR:
	    return res;
	case DB_INFO:
	    throw Exception(EXLOC, "Invalid result type");
	}
    }

    int col=1;
    CegoField *pF = schema.First();
    while ( pF ) 
    {
	Chain colName = pF->getAttrName();
	Chain colPos = Chain(NETMNG_COLPREFIX) + Chain(col);
	
	if ( (*_pRow)->hasAttribute(colPos) == false )
	{
	    CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, NULL_TYPE, 0));	    
	    CegoFieldValue fv;
	    f.setValue(fv);
	    fvl.Insert(f);	    
	}
	else
	{
	
	    Chain colVal = (*_pRow)->getAttributeValue(colPos);

	    if ( pF->getType() == INT_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, INT_TYPE, sizeof(int)));	    
		CegoFieldValue fv(INT_TYPE, colVal);	    
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == LONG_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, LONG_TYPE, sizeof(long)));	    
		CegoFieldValue fv(LONG_TYPE, colVal);	    
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == VARCHAR_TYPE )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, VARCHAR_TYPE, colVal.length()));
		
		char *pS = (char*)malloc(colVal.length());
		strcpy(pS, (char*)colVal);
		
		CegoFieldValue fv(VARCHAR_TYPE, pS, colVal.length(), true);
		
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == BOOL_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, BOOL_TYPE, sizeof(char)));	
		CegoFieldValue fv(BOOL_TYPE, colVal);
		
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == DATETIME_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, DATETIME_TYPE, sizeof(int)));		    
		CegoFieldValue fv(DATETIME_TYPE, colVal);	  
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == FLOAT_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, FLOAT_TYPE, sizeof(float)));		    
		CegoFieldValue fv(FLOAT_TYPE, colVal);	  
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == DOUBLE_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, DOUBLE_TYPE, sizeof(double)));		    
		CegoFieldValue fv(DOUBLE_TYPE, colVal);	  
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == SMALLINT_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, SMALLINT_TYPE, sizeof(short)));		    
		CegoFieldValue fv(SMALLINT_TYPE, colVal);	  
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == TINYINT_TYPE  )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, TINYINT_TYPE, sizeof(char)));		    
		CegoFieldValue fv(TINYINT_TYPE, colVal);	  
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == BIGINT_TYPE )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, BIGINT_TYPE, colVal.length()));
		CegoFieldValue fv(BIGINT_TYPE, colVal);	    
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == DECIMAL_TYPE )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, DECIMAL_TYPE, colVal.length()));
		CegoFieldValue fv(DECIMAL_TYPE, colVal);	    
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == FIXED_TYPE )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, FIXED_TYPE, colVal.length()));
		CegoFieldValue fv(FIXED_TYPE, colVal);	    
		f.setValue(fv);
		fvl.Insert(f);
	    }
	    else if ( pF->getType() == BLOB_TYPE )
	    {
		CegoField f(CegoField(pF->getTableName(), pF->getTableAlias(), colName, BLOB_TYPE, 2 * sizeof(int)));
		CegoFieldValue fv(BLOB_TYPE, colVal);	    
		f.setValue(fv);
		fvl.Insert(f);
	    }
	}
	
	col++;
	pF = schema.Next();
    }
    return DB_DATA;
}

void CegoDbHandler::abortQuery()
{
    _pModule->log(_modId, Logger::DEBUG, Chain("Aborting query"));   
    _pN->sendChar(QUERY_ABORT);

    _pN->readMsg();

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    
    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, _pN->getMsg());
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();

    if ( docType == Chain(XML_OK_DOC) )
    {
	_rowList.Empty();
	_pRow = 0;
    }
    else if ( docType == Chain(XML_ERROR_DOC) )
    {
	_rowList.Empty();
	_pRow = 0;
    }
}

void CegoDbHandler::resetQuery()
{
    _pModule->log(_modId, Logger::DEBUG, Chain("Resetting query"));   
    _pN->sendChar(QUERY_RESET);

    _rowList.Empty();
    _pRow = 0;

}

void CegoDbHandler::sendCollectedData()
{
    
    char c = _pN->recvChar();
    
    if ( c == QUERY_ABORT )
    {
	_pModule->log(_modId, Logger::LOGERR, Chain("User query abort"));
	throw Exception(EXLOC, "Query aborted by user");
    }
    else if ( c == QUERY_RESET )
    {
	_wasReset=true;
	_xml.getDocument()->clear();
    }
    else if ( c == QUERY_MOREDATA )
    {
	Chain xmlString;
	
	_xml.getDocument()->setDocType(XML_DATA_DOC);
	_xml.getXMLChain(xmlString);

	_pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
	_pModule->log(_modId, Logger::DEBUG, xmlString);
	_pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));
        
	_pN->setMsg(xmlString, xmlString.length());
        
	_pN->writeMsg();
	
	_xml.getDocument()->clear();
    }
    else
    {
	throw Exception(EXLOC, "Unknown query sync");
    }

    return;
}

bool CegoDbHandler::wasReset()
{
    if ( _wasReset )
    {
	_wasReset = false;
	return true;
    }
    else
    {
	_wasReset = false;
	return false;
    }
    
}
void CegoDbHandler::sendFinishData()
{


    char c = _pN->recvChar();
    
    if ( c == QUERY_ABORT )
    {
	// can be ignored
	// _pModule->log(_modId, Logger::LOGERR, Chain("User query abort"));
	// throw Exception(EXLOC, "Query aborted by user");
    }
    else if ( c == QUERY_RESET )
    {
	// _wasReset=true;
	// _xml.getDocument()->clear();
    }
    else if ( c == QUERY_MOREDATA )
    {
	// nothing to to
    }

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    _xml.getDocument()->setRootElement(pRoot);
    
    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());
    
    _pN->writeMsg();
}

void CegoDbHandler::sendErrorData(const Chain& msg)
{

    char c = _pN->recvChar();
    
    if ( c == QUERY_ABORT )
    {
	// can be ignored
	// we send error back anyway
    }
    else if ( c == QUERY_RESET )
    {
	// we send error back anyway
    }
    else if ( c == QUERY_MOREDATA )
    {
	// we send error back anyway
	
    }

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_ERROR_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_MSG_ATTR, msg);
    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());
    
    _pN->writeMsg();
}


void CegoDbHandler::sendSessionConfirm(const Chain& msg, 
				       long tid, 
				       const Chain& dbName, 
				       const Chain& dbVersion)
{
    
    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_MSG_ATTR, msg);
    pRoot->setAttribute(XML_TID_ATTR, Chain(tid));
    pRoot->setAttribute(XML_DBPRODNAME_ATTR, dbName);
    pRoot->setAttribute(XML_DBPRODVERSION_ATTR, dbVersion);

    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());

    _pN->writeMsg();
    
}



void CegoDbHandler::sendResponse(const Chain& msg, long affected)
{

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_MSG_ATTR, msg);
    pRoot->setAttribute(XML_AFFECTED_ATTR, Chain(affected));

    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());

    _pN->writeMsg();
    
}

void CegoDbHandler::sendError(const Chain& msg)
{

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_ERROR_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_MSG_ATTR, msg);

    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());
    
    _pN->writeMsg();
    
}

void CegoDbHandler::sendBlobInfo(int fileId, int pageId)
{

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_FILEID_ATTR, Chain(fileId));
    pRoot->setAttribute(XML_PAGEID_ATTR, Chain(pageId));

    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());

    _pN->writeMsg();
    
}

void CegoDbHandler::sendBlobSize(long blobSize)
{

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_SIZE_ATTR, Chain(blobSize));

    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());

    _pN->writeMsg();
    
}


CegoDbHandler::ResultType CegoDbHandler::putBlob(const Chain& tableSet, CegoBlob& blob)
{

    _xml.getDocument()->clear();
    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_TABLESET_ATTR, tableSet);
    pRoot->setAttribute(XML_SIZE_ATTR, blob.getSize());

    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(XML_PUTBLOB_REQUEST);
    
    Chain request;
    _xml.getXMLChain(request);
    
    _pN->setMsg(request, request.length());
    
    _pN->writeMsg();
    
    _pN->readMsg();

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();

    if ( docType == Chain(XML_ERROR_DOC) )
    {
	return DB_ERROR;
    }    

    pRoot = _xml.getDocument()->getRootElement();
    if ( pRoot )
    {
	blob.setFileId(pRoot->getAttributeValue(XML_FILEID_ATTR).asInteger());
	blob.setPageId(pRoot->getAttributeValue(XML_PAGEID_ATTR).asInteger());
    }

    blob.reset();
    
    while ( blob.nextChunk(BLOBCHUNKSIZE) )
    {
	
	_pN->setMsg((char*)blob.getChunkPtr(), blob.getChunkSize());
	
	_pN->writeMsg();
	
	if ( _pN->recvAck() == false )
	{
	    _pModule->log(_modId, Logger::LOGERR, Chain("User query abort"));
	    return DB_ERROR;
	}
    }
    
    return DB_OK;
}

CegoDbHandler::ResultType CegoDbHandler::getBlob(const Chain& tableSet, CegoBlob& blob)
{

    _xml.getDocument()->clear();
    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_TABLESET_ATTR, tableSet);
    pRoot->setAttribute(XML_FILEID_ATTR, blob.getFileId());
    pRoot->setAttribute(XML_PAGEID_ATTR, blob.getPageId());

    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(XML_GETBLOB_REQUEST);
    
    Chain request;
    _xml.getXMLChain(request);
    
    _pN->setMsg(request, request.length());
    
    _pN->writeMsg();
    
    _pN->readMsg();

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();

    if ( docType == Chain(XML_ERROR_DOC) )
    {
	return DB_ERROR;
    }    

    long blobSize;

    pRoot = _xml.getDocument()->getRootElement();
    if ( pRoot )
    {
	blobSize = pRoot->getAttributeValue(XML_SIZE_ATTR).asLong();
    }

    blob.allocate(blobSize);
    blob.reset();
    
    int recvSize = 0;
    while ( recvSize < blobSize )
    {
       	_pN->sendAck();
	_pN->readMsg();
	Chain msg(_pN->getMsg(), _pN->getMsgSize() );

	blob.putChunk((unsigned char*)_pN->getMsg(), _pN->getMsgSize() );

	recvSize += _pN->getMsgSize();
    }
    
    return DB_OK;
}

CegoDbHandler::ResultType CegoDbHandler::delBlob(const Chain& tableSet, CegoBlob& blob)
{

    _xml.getDocument()->clear();
    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_TABLESET_ATTR, tableSet);
    pRoot->setAttribute(XML_FILEID_ATTR, blob.getFileId());
    pRoot->setAttribute(XML_PAGEID_ATTR, blob.getPageId());

    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(XML_DELBLOB_REQUEST);
    
    Chain request;
    _xml.getXMLChain(request);
    
    _pN->setMsg(request, request.length());
    
    _pN->writeMsg();
    
    _pN->readMsg();

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();

    if ( docType == Chain(XML_OK_DOC) )
    {
	return DB_OK;
    }	
    else if ( docType == Chain(XML_ERROR_DOC) )
    {
	return DB_ERROR;
    }
    else
    {
	throw Exception(EXLOC, Chain("Invalid document type"));
    }
}

void CegoDbHandler::sendProcResult(const Chain& msg, const ListT<CegoProcVar>& outParamList, CegoFieldValue* pRetValue)
{
    _pModule->log(_modId, Logger::DEBUG, Chain("Sending procedure result"));

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_MSG_ATTR, msg);

    if ( pRetValue )
    {
	Element *pOutParam = new Element(XML_OUTPARAM_ELEMENT);
	
	CegoXMLHelper xh;

	CegoTypeConverter tc;
	pOutParam->setAttribute(XML_TYPE_ATTR, tc.getTypeString(pRetValue->getType()));

	pOutParam->setAttribute(XML_VALUE_ATTR, pRetValue->valAsChain());
	pRoot->addContent(pOutParam);
    }
    
    CegoProcVar *pVar = outParamList.First();
    while ( pVar )
    {
	Element *pOutParam = new Element(XML_OUTPARAM_ELEMENT);
	pOutParam->setAttribute(XML_NAME_ATTR, pVar->getName());
	
	CegoTypeConverter tc;
	pOutParam->setAttribute(XML_TYPE_ATTR, tc.getTypeString(pVar->getValue().getType()));
	pOutParam->setAttribute(XML_VALUE_ATTR, pVar->getValue().valAsChain());
	
	pVar = outParamList.Next();
	pRoot->addContent(pOutParam);
    }
   
    _xml.getDocument()->setRootElement(pRoot);

    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());

    _pN->writeMsg();
    
}

void CegoDbHandler::sendProdInfo()
{

    _xml.getDocument()->clear();
    _xml.getDocument()->setDocType(XML_OK_DOC);

    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_DBPRODNAME_ATTR, XML_DBPRODNAME_VALUE);
    pRoot->setAttribute(XML_DBPRODVERSION_ATTR, XML_DBPRODVERSION_VALUE);

    _xml.getDocument()->setRootElement(pRoot);


    Chain xmlString;
    _xml.getXMLChain(xmlString);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, xmlString);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _pN->setMsg(xmlString, xmlString.length());

    _pN->writeMsg();
    
}

CegoDbHandler::ResultType CegoDbHandler::getMoreTableData()
{

    _pModule->log(_modId, Logger::DEBUG, Chain("Sending ack ..."));
    
    _pN->sendChar(QUERY_MOREDATA);

    _pModule->log(_modId, Logger::DEBUG, Chain("Reading data ..."));

    _pN->readMsg();

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, _pN->getMsg());
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();

    if ( docType == Chain(XML_DATA_DOC) )
    {
	Element *pRoot = _xml.getDocument()->getRootElement();
	if ( pRoot )
	{
	    _rowList = pRoot->getChildren(XML_ROW_ELEMENT);
	}
	return DB_DATA;
    }
    else if ( docType == Chain(XML_OK_DOC) )
    {
	_rowList.Empty();
	return DB_OK;
    }
    else if ( docType == Chain(XML_ERROR_DOC) )
    {
	return DB_ERROR;
    }
    else
    {
	throw Exception(EXLOC, Chain("Invalid document type"));
    }
}

CegoDbHandler::RequestType CegoDbHandler::acceptRequest()
{
    _pModule->log(_modId, Logger::DEBUG, Chain("Accepting request ..."));
    
    if ( _pN->waitMsg(NETMNG_WAITMSG_TIMEOUT) == false )
	return CegoDbHandler::REQTIMEOUT;

    _pN->readMsg();

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, _pN->getMsg());
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    
    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    _xml.parse();
        
    Chain docType = _xml.getDocument()->getDocType();
    
    if ( docType == Chain(XML_QUERY_REQUEST) ) 
    {
	return CegoDbHandler::QUERY;
    }
    else if ( docType == Chain(XML_QUERYABORT_REQUEST) ) 
    {
	return CegoDbHandler::QUERYABORT;
    }
    else if ( docType == Chain(XML_DBPRODINFO_REQUEST) ) 
    {
	return CegoDbHandler::DBPRODINFO;
    }
    else if (  docType == Chain(XML_CREATETABLE_REQUEST) ) 
    {
	return CegoDbHandler::CREATETABLE;
    }
    else if (  docType == Chain(XML_CREATEVIEW_REQUEST) ) 
    {
	return CegoDbHandler::CREATEVIEW;
    }
    else if (  docType == Chain(XML_CREATEPROCEDURE_REQUEST) ) 
    {
	return CegoDbHandler::CREATEPROCEDURE;
    }
    else if (  docType == Chain(XML_CREATECHECK_REQUEST) ) 
    {
	return CegoDbHandler::CREATECHECK;
    }
    else if (  docType == Chain(XML_ALTERTABLE_REQUEST) ) 
    {
	return CegoDbHandler::ALTERTABLE;
    }
    else if (  docType == Chain(XML_CREATEINDEX_REQUEST) ) 
    {
	return CegoDbHandler::CREATEINDEX;
    }
    else if (  docType == Chain(XML_INSERT_REQUEST) ) 
    {
	return CegoDbHandler::INSERT;
    }
    else if (  docType == Chain(XML_DELETE_REQUEST) ) 
    {
	return CegoDbHandler::DELETE;
    }
    else if (  docType == Chain(XML_UPDATE_REQUEST) ) 
    {
	return CegoDbHandler::UPDATE;
    }
    else if (  docType == Chain(XML_OBJECTINFO_REQUEST) ) 
    {
	return CegoDbHandler::OBJECTINFO;
    }
    else if (  docType == Chain(XML_DROP_OBJECT_REQUEST) ) 
    {
	return CegoDbHandler::DROPOBJECT;
    }
    else if (  docType == Chain(XML_GETTABLE_REQUEST) ) 
    {
	return CegoDbHandler::GETTABLE;
    }
    else if (  docType == Chain(XML_GETOBJLIST_REQUEST) ) 
    {
	return CegoDbHandler::GETOBJLIST;
    }
    else if (  docType == Chain(XML_GETOBJLISTBYTABLE_REQUEST) ) 
    {
	return CegoDbHandler::GETOBJLISTBYTABLE;
    }
    else if (  docType == Chain(XML_OBJRENAME_REQUEST) ) 
    {
	return CegoDbHandler::OBJRENAME;
    }
    else if (  docType == Chain(XML_REORG_REQUEST) ) 
    {
	return CegoDbHandler::REORGOBJ;
    }
    else if (  docType == Chain(XML_SYNC_REQUEST) ) 
    {
	return CegoDbHandler::SYNC;
    }
    else if (  docType == Chain(XML_GETPAGECOUNT_REQUEST) ) 
    {
	return CegoDbHandler::GETPAGECOUNT;
    }
    else if (  docType == Chain(XML_PUTBLOB_REQUEST) ) 
    {
	return CegoDbHandler::PUTBLOB;
    }
    else if (  docType == Chain(XML_GETBLOB_REQUEST) ) 
    {
	return CegoDbHandler::GETBLOB;
    }
    else if (  docType == Chain(XML_DELBLOB_REQUEST) ) 
    {
	return CegoDbHandler::DELBLOB;
    }
    else if (  docType == Chain(XML_STARTTRANSACTION_REQUEST) ) 
    {
	return CegoDbHandler::STARTTRANSACTION;
    }
    else if (  docType == Chain(XML_COMMITTRANSACTION_REQUEST) ) 
    {
	return CegoDbHandler::COMMITTRANSACTION;
    }
    else if (  docType == Chain(XML_ROLLBACKTRANSACTION_REQUEST) ) 
    {
	return CegoDbHandler::ROLLBACKTRANSACTION;
    }
    else if (  docType == Chain(XML_GETTID_REQUEST) ) 
    {
	return CegoDbHandler::GETTID;
    }
    else if (  docType == Chain(XML_SESSION_CLOSE) ) 
    {
	return CegoDbHandler::SESSION_CLOSE;
    }
    else
    {
	return CegoDbHandler::UNKNOWN;
    } 
    
}

void CegoDbHandler::getQueryArg(Chain& cmd)
{

    Element *pRoot = _xml.getDocument()->getRootElement();
    
    if ( pRoot )
    {
	cmd = pRoot->getAttributeValue(XML_CMD_ATTR);
    }	
}

void CegoDbHandler::sendSchema(const ListT<CegoField>& schema, const Chain& format)
{

    _pModule->log(_modId, Logger::DEBUG, Chain("Sending schema ..."));

    _xml.getDocument()->clear();
    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->setAttribute(XML_FORMAT_ATTR, format);

    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(XML_DATA_DOC);

    CegoField *pF = schema.First();
    while ( pF ) 
    {

	Chain tname;
	if (pF->getTableAlias().length() > 0)
	{
	    tname = pF->getTableAlias();
	}
	else
	{
	    tname = pF->getTableName();
	}
	    
	Element *pColElement = new Element(XML_SCHEMA_ELEMENT);
	pColElement->setAttribute(XML_TABLENAME_ATTR, tname);

	CegoXMLHelper xh;
	xh.setColInfo(pColElement, pF);
	
	pRoot->addContent(pColElement);
	pF = schema.Next();
    }

    Chain response;
    _xml.getXMLChain(response);


    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, response);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));


    _pN->setMsg((char*)response, response.length());
    _pN->writeMsg();

    // document must be cleared because of follow up collectData calls 
    _xml.getDocument()->clear();
       
}

void CegoDbHandler::sendObjInfo(CegoDecodableObject& oe)
{

    _xml.getDocument()->clear();
    
    Element* pRoot = new Element(XML_FRAME_ELEMENT);
    pRoot->addContent(oe.getElement());
    
    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(XML_INFO_DOC);

    Chain response;
    _xml.getXMLChain(response);

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, response);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));
    
    _pN->setMsg((char*)response, response.length());
    _pN->writeMsg();

    _xml.getDocument()->clear();
       
}

void CegoDbHandler::collectData(const ListT<CegoField>& schema)
{

    Element* pRoot = _xml.getDocument()->getRootElement();
    if ( pRoot == 0)
    {
	pRoot = new Element(XML_FRAME_ELEMENT);
	_xml.getDocument()->setRootElement(pRoot);	
    }
    
    Element *pRowElement = new Element(XML_ROW_ELEMENT);

    int col=1;
    CegoField *pF = schema.First();
    while ( pF )
    {

	Chain colPos = Chain(NETMNG_COLPREFIX) + Chain(col);

	if ( ! pF->getValue().isNull() )
	{
	    pRowElement->setAttribute(colPos, pF->getValue().valAsChain());   
	}
	col++;
	pF = schema.Next();
    }
    
    pRoot->addContent(pRowElement);

}

void CegoDbHandler::collectData(const ListT<CegoField>& schema, const ListT<CegoFieldValue>& fvl)
{

    Element* pRoot = _xml.getDocument()->getRootElement();
    if ( pRoot == 0)
    {
	pRoot = new Element(XML_FRAME_ELEMENT);
	_xml.getDocument()->setRootElement(pRoot);	
    }
    
    Element *pRowElement = new Element(XML_ROW_ELEMENT);

    int col=1;
    CegoField *pF = schema.First();
    CegoFieldValue *pFV = fvl.First();

    while ( pF && pFV )
    {
	
	Chain colPos = Chain(NETMNG_COLPREFIX) + Chain(col);

	if ( ! pFV->isNull() )
	{
	    pRowElement->setAttribute(colPos, pFV->valAsChain());
	}
	col++;
	pF = schema.Next();
	pFV = fvl.Next();
    }
    
    pRoot->addContent(pRowElement);    
}

NetHandler* CegoDbHandler::getNetHandler()
{
    return _pN;
}


CegoDbHandler::ResultType CegoDbHandler::sendReq(const Chain& reqType, Element* pRoot)
{
    _xml.getDocument()->clear();
    _xml.getDocument()->setRootElement(pRoot);
    _xml.getDocument()->setDocType(reqType);
    
    Chain request;
    _xml.getXMLChain(request);
    _xml.getDocument()->clear();


    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, request);
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));
    
    _pN->setMsg(request, request.length());

    _pN->writeMsg();
    
    _pN->readMsg();

    _pModule->log(_modId, Logger::DEBUG, Chain("--- XML ---"));
    _pModule->log(_modId, Logger::DEBUG, _pN->getMsg());
    _pModule->log(_modId, Logger::DEBUG, Chain("--- --- ---"));

    _xml.getDocument()->clear();
    _xml.setChain( _pN->getMsg() );
    _xml.parse();

    Chain docType = _xml.getDocument()->getDocType();
    
    if ( docType == Chain(XML_OK_DOC) )
    {
	return DB_OK;
    }
    else if ( docType == Chain(XML_ERROR_DOC) )
    {
	return DB_ERROR;
    }
    else if ( docType == Chain(XML_INFO_DOC) )
    {
	return DB_INFO;
    }
    else
    {
	throw Exception(EXLOC, Chain("Invalid document type"));
    }
}
