///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoTransactionManager.cc
// -------------------------
// Cego transaction manager implementation
//     
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoTransactionManager
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _REENTRANT
#define _REENTRANT    /* basic 3-lines for threads */
#endif

// base includes
#include <lfc/Exception.h>
#include <lfc/Sleeper.h>
#include <lfc/ThreadLock.h>

// cego includes
#include "CegoTransactionManager.h"

// system includes
#include <string.h>
#include <stdlib.h>

#include "CegoDataType.h"
#include "CegoTupleState.h"
#include "CegoTableManager.h"
#include "CegoIndexManager.h"

#define SYS_RBCATLOG "rbcatlog"
#define SYS_UPDTAB "updtab"

#define SYS_RBCATLOG_TID "tid"
#define SYS_RBCATLOG_FILEID "fileid"
#define SYS_RBCATLOG_PAGEID "pageid"
#define SYS_RBCATLOG_OFFSET "offset"
#define SYS_RBCATLOG_TABLE "table"
#define SYS_RBCATLOG_RB_FILEID "rb_fileid"
#define SYS_RBCATLOG_RB_PAGEID "rb_pageid"
#define SYS_RBCATLOG_RB_OFFSET "rb_offset"

#define SYS_UPDTAB_TID "tid"
#define SYS_UPDTAB_FILEID "fileid"
#define SYS_UPDTAB_PAGEID "pageid"
#define SYS_UPDTAB_OFFSET "offset"

static ThreadLock taLock[TABMNG_MAXTABSET];
extern bool __lockStatOn;

CegoTransactionManager::CegoTransactionManager(CegoTableManager *pTM)
{
    _pTM = pTM;

    CegoFieldValue defVal;
    _rbcatSchema.Insert(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_TID, INT_TYPE, sizeof(int), defVal, true, 1));
    _rbcatSchema.Insert(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_FILEID, INT_TYPE, sizeof(int), defVal, true, 2));
    _rbcatSchema.Insert(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_PAGEID, INT_TYPE, sizeof(int), defVal, true, 3));
    _rbcatSchema.Insert(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_OFFSET, INT_TYPE, sizeof(int), defVal, true, 4));
    _rbcatSchema.Insert(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_TABLE, VARCHAR_TYPE, MAX_OBJNAME_LEN, defVal, false, 5));
    
    _updSchema.Insert(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_TID, INT_TYPE, sizeof(int), defVal, true, 1));
    _updSchema.Insert(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_FILEID, INT_TYPE, sizeof(int), defVal, true, 2));
    _updSchema.Insert(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_PAGEID, INT_TYPE, sizeof(int), defVal, true, 3));
    _updSchema.Insert(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_OFFSET, INT_TYPE, sizeof(int), defVal, true, 4));

    _pDBMng = pTM->getDBMng();

    _modId = _pDBMng->getModId("CegoTransactionManager");
}

CegoTransactionManager::~CegoTransactionManager()
{
    for ( int i=0; i < TABMNG_MAXTABSET; i++)
	if ( _pRBPage[i].isFixed() )
	{
	    _pDBMng->bufferUnfix(_pRBPage[i], true, _pTM->getLockHandler());
	}
}

void CegoTransactionManager::release(int tabSetId)
{
    if ( _pRBPage[tabSetId].isFixed() )
    {
	_pDBMng->bufferUnfix(_pRBPage[tabSetId], true, _pTM->getLockHandler());
    }
}

void CegoTransactionManager::reorgSystemSpace(int tabSetId)
{

    taLock[tabSetId].writeLock(TS_LOCKTIMEOUT);

    try
    {
	_pTM->reorgObject(tabSetId, SYS_RBCATLOG, CegoObject::RBSEG);
	_pTM->reorgObject(tabSetId, SYS_UPDTAB, CegoObject::RBSEG);

    }
    catch ( Exception e )
    {
	taLock[tabSetId].unlock();
	throw Exception(EXLOC, Chain("Reorganization failed"), e);
    }
    taLock[tabSetId].unlock();
    
}

void CegoTransactionManager::initLock(int tabSetId)
{    
    taLock[tabSetId].init(LCKMNG_LOCKWAITDELAY, __lockStatOn);
    Chain tableSet = _pTM->getDBMng()->getTabSetName(tabSetId);
    taLock[tabSetId].setId(tableSet + Chain("-TA"));
}

void CegoTransactionManager::createSystemSpace(int tabSetId)
{    
    _pTM->createDataTable(tabSetId, SYS_UPDTAB, CegoObject::RBSEG, _updSchema);
    _pTM->createDataTable(tabSetId, SYS_RBCATLOG, CegoObject::RBSEG, _rbcatSchema);   
}

void CegoTransactionManager::newRBEntry(int tabSetId, int tid, int fileId, int pageId, int offset, const Chain& tableName)
{

    if ( _pRBPage[tabSetId].isFixed() == false )
    {
	_pTM->getObjectWithFix(tabSetId, SYS_RBCATLOG, CegoObject::RBSEG, _rbe[tabSetId], _pRBPage[tabSetId]);
	_rbEntry[tabSetId] = CegoDataPointer(_pRBPage[tabSetId].getFileId(), _pRBPage[tabSetId].getPageId(), _pRBPage[tabSetId].getEntryPos());
    }
    
    CegoFieldValue fv1(INT_TYPE, &tid, sizeof(int));
    CegoField f1(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_TID, INT_TYPE, sizeof(int), fv1, false, 1));
    
    CegoFieldValue fv2(INT_TYPE, &fileId, sizeof(int));
    CegoField f2(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_FILEID, INT_TYPE, sizeof(int), fv2, true, 2));	
    
    CegoFieldValue fv3(INT_TYPE, &pageId, sizeof(int));
    CegoField f3(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_PAGEID, INT_TYPE, sizeof(int), fv3, true, 3));	
    

    CegoFieldValue fv4(INT_TYPE, &offset, sizeof(int));
    CegoField f4(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_OFFSET, INT_TYPE, sizeof(int), fv4, true, 4));
    
    CegoFieldValue fv5(VARCHAR_TYPE, (char*)tableName, tableName.length());
    CegoField f5(CegoField(SYS_RBCATLOG, SYS_RBCATLOG, SYS_RBCATLOG_TABLE, VARCHAR_TYPE, MAX_OBJNAME_LEN, fv5, false, 5));	
    
    ListT<CegoField> fl;
    fl.Insert(f1);
    fl.Insert(f2);
    fl.Insert(f3);
    fl.Insert(f4);
    fl.Insert(f5);
    
    
    int rbtid = 0;
    int rbtastep = 0;
    CegoTupleState ts = COMMITTED;
    
    char *pBufBase = 0;
    int buflen = 0;
    
    _qh.encodeFVL(rbtid, rbtastep, ts, fl, pBufBase, buflen);
    
    try
    {	
	taLock[tabSetId].readLock(TS_LOCKTIMEOUT);	
	CegoDataPointer dp = _pTM->insertData(_rbEntry[tabSetId], _rbe[tabSetId], pBufBase, buflen, true);
    }
    catch ( Exception e )
    {
	taLock[tabSetId].unlock();
	free (pBufBase);
	throw Exception(EXLOC, Chain("Cannot make rollback entry"), e);
    }
    taLock[tabSetId].unlock();
    free (pBufBase);

}

void CegoTransactionManager::getTransactionAffectedTables(int tabSetId, int tid, SetT<Chain>& tableList)
{


    taLock[tabSetId].readLock(TS_LOCKTIMEOUT);

    try
    {
		
	CegoObjectCursor *pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);
	
	ListT<CegoField> schema = _rbcatSchema;

	try
	{
	    
	    CegoDataPointer dp;
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	    	    
	    while (moreTuple)
	    {
		
		CegoField *pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TID));
		if (pF)
		{
		    int val;
		    memcpy(&val, pF->getValue().getValue(), sizeof(int));
		    if (val == tid)
		    {
			Chain tableName;
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TABLE));
			if (pF)
			{
			    tableName = Chain((char*)pF->getValue().getValue());			
			    tableList.Insert(tableName);			      
			}
		    }
		}
		
		moreTuple = _pTM->getNextTuple(pOC, schema, dp);
	    }
	    
	}
	
	catch ( Exception e )
	{
	    pOC->abort();
	    delete pOC;
	    throw Exception(EXLOC, Chain("Cannot get affected tables"), e);
	}	
	
	pOC->abort();
	delete pOC;

    }
    catch ( Exception e )
    {
	taLock[tabSetId].unlock();
	throw Exception(EXLOC, Chain("Cannot get affected tables"), e);
    }
    
    taLock[tabSetId].unlock();

}

long CegoTransactionManager::commitTransaction(int tabSetId, int tid)
{

    long opCount = 0;

    taLock[tabSetId].readLock(TS_LOCKTIMEOUT);

    try
    {
		
	CegoObjectCursor *pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);
	
	ListT<CegoField> schema = _rbcatSchema;
	
	// Step 1 : Committing all attached tuples
	
	try  {
	    
	    
	    CegoDataPointer dp;
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	    
	    
	    Chain cachedTable;
	    ListT<CegoField> cachedFvl;
	    ListT<CegoTableObject> cachedIdxList;
	    ListT<CegoKeyObject> cachedKeyList;
	    ListT<CegoCheckObject> cachedCheckList;
	    
	    
	    while (moreTuple)
	    {
		
		CegoField *pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TID));
		if (pF)
		{
		    int val;
		    memcpy(&val, pF->getValue().getValue(), sizeof(int));
		    if (val == tid)
		    // if (*(int*)(pF->getValue().getValue()) == tid)
		    {
			
			int fileId, pageId, offset;
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_FILEID));
			if (pF)
			    memcpy(&fileId, pF->getValue().getValue(), sizeof(int));
			    // fileId = *(int*)pF->getValue().getValue();
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_PAGEID));
			if (pF)
			    memcpy(&pageId, pF->getValue().getValue(), sizeof(int));
			    // pageId = *(int*)pF->getValue().getValue();
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_OFFSET));
			if (pF)
			    memcpy(&offset, pF->getValue().getValue(), sizeof(int));
			    // offset = *(int*)pF->getValue().getValue();
			
			CegoDataPointer dp(fileId, pageId, offset);
			
			int tid;
			int tastep;
			CegoTupleState ts;
			
			_pTM->getTupleInfo(tabSetId, dp, tid, tastep, ts);			
			_pTM->setTupleInfo(tabSetId, dp, 0, 0, COMMITTED);
			opCount++;
			if ( ts == DELETED || ts == OBSOLETE )
			{
			    
			    Chain tableName;
			    
			    pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TABLE));
			    if (pF)
				tableName = Chain((char*)pF->getValue().getValue());
			    
			    
			    if ( cachedTable != tableName )
			    {
				CegoTableObject oe;
				_pTM->getObject(tabSetId, tableName, CegoObject::TABLE, oe);
				cachedFvl = oe.getSchema();
				cachedTable = tableName;
				cachedIdxList.Empty();
				cachedKeyList.Empty();
				cachedCheckList.Empty();
				_pTM->getObjectListByTable(tabSetId, cachedTable, cachedIdxList, cachedKeyList, cachedCheckList);
				
			    }
			    
			    char* p;
			    int len;
			    unsigned long lockId;
			    
			    try
			    {
				lockId = _pTM->claimDataPtr(tabSetId, CegoLockHandler::READ, CegoBufferPool::SYNC, dp, p, len);
				
				int tid;
				int tastep;
				CegoTupleState ts;
				_qh.decodeFVL(cachedFvl, p, len, tid, tastep, ts);
				
				_pTM->deleteDataTable(tabSetId, cachedTable, CegoObject::TABLE, dp, cachedFvl, cachedIdxList, cachedKeyList, false);
				
			    }
			    catch ( Exception e)
			    {   
				_pTM->releaseDataPtr(lockId);
				throw Exception(EXLOC, Chain("Cannot commit transaction"), e );
			    }
			    _pTM->releaseDataPtr(lockId);		    
			} 
		    }
		}
		else
		{
		    throw Exception(EXLOC, "Cannot get tid entry");
		}
		
		moreTuple = _pTM->getNextTuple(pOC, schema, dp);
	    }
	}
	catch ( Exception e )
	{
	    pOC->abort();
	    delete pOC;
	    throw Exception(EXLOC, Chain("Cannot commit transaction"), e);
	}
	
	
	pOC->abort();
	delete pOC;
	
	// Step 2 : Cleaning up rollback segment
	
	pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);
	
	try  {
	    
	    CegoDataPointer dp;	
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	    
	    while (moreTuple)
	    {
		
		CegoField *pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TID));
		if (pF)
		{
		    if (*(int*)(pF->getValue().getValue()) == tid)
		    {
			CegoDataPointer ddp = dp; 
			moreTuple = _pTM->getNextTuple(pOC, schema, dp);
			
			_pTM->deleteData(CegoObject::RBSEG, tabSetId, ddp);
	    	    	
		    }
		    else
		    {
			moreTuple = _pTM->getNextTuple(pOC, schema, dp);
		    }
		}
		else
		{
		    throw Exception(EXLOC, "Cannot get tid entry");
		}
	    }
	}
	catch ( Exception e )
	{
	    if ( pOC )
	    {
		pOC->abort();
		delete pOC;
	    }
	    throw Exception(EXLOC, Chain("Cannot commit transaction"), e);
	}
	
	pOC->abort();
	delete pOC;
	
    }
    catch ( Exception e )
    {
	taLock[tabSetId].unlock();
	throw Exception(EXLOC, Chain("Cannot commit transaction"), e);
    }

    taLock[tabSetId].unlock();

    return opCount;
}

long CegoTransactionManager::rollbackTransaction(int tabSetId, int tid)
{
    
    long opCount = 0;

    taLock[tabSetId].readLock(TS_LOCKTIMEOUT);	

    try
    {
	
	CegoObjectCursor *pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);
	
	ListT<CegoField> schema = _rbcatSchema;
	
	// step 1 ( tid setting )
	
	try 
	{
	    
	    CegoDataPointer dp;
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	    
	    Chain cachedTable;
	    ListT<CegoField> cachedFvl;
	    ListT<CegoTableObject> cachedIdxList;
	    ListT<CegoKeyObject> cachedKeyList;
	    ListT<CegoCheckObject> cachedCheckList;
	    
	    while (moreTuple)
	    {
		CegoField *pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TID));
		if (pF)
		{

		    // if (*(int*)(pF->getValue().getValue()) == tid)

		    int val;
		    memcpy(&val, pF->getValue().getValue(), sizeof(int));
		    if (val == tid)
		    {
			
			int fileId, pageId, offset;
			Chain tableName;
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_FILEID));
			if (pF)
			    memcpy(&fileId, pF->getValue().getValue(), sizeof(int));
			    // fileId = *(int*)pF->getValue().getValue();
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_PAGEID));
			if (pF)
			    memcpy(&pageId, pF->getValue().getValue(), sizeof(int));
			    // pageId = *(int*)pF->getValue().getValue();
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_OFFSET));
			if (pF)
			    memcpy(&offset, pF->getValue().getValue(), sizeof(int));
			    // offset = *(int*)pF->getValue().getValue();
			
			pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TABLE));
			if (pF)
			    tableName = Chain((char*)pF->getValue().getValue());
			
			
			CegoDataPointer dp(fileId, pageId, offset);
			
			int tid;
			int tastep;
			CegoTupleState ts;
			
			_pTM->getTupleInfo(tabSetId, dp, tid, tastep, ts);			
			_pTM->setTupleInfo(tabSetId, dp, 0, 0, COMMITTED);
			opCount++;

			if ( ts == INSERTED || ts == OBSOLETE ) 
			{
			    
			    if ( cachedTable != tableName )
			    {
				CegoTableObject oe;
				_pTM->getObject(tabSetId, tableName, CegoObject::TABLE, oe);
				cachedFvl = oe.getSchema();
				cachedTable = tableName;
				cachedIdxList.Empty();
				cachedKeyList.Empty();
				cachedCheckList.Empty();
				_pTM->getObjectListByTable(tabSetId, cachedTable, cachedIdxList, cachedKeyList, cachedCheckList);
				
			    }
			    			    
			    char* p;
			    int len;
			    unsigned long lockId;
			    
			    try
			    {
				lockId = _pTM->claimDataPtr(tabSetId, CegoLockHandler::READ, CegoBufferPool::SYNC, dp, p, len);
				
				int tid;
				CegoTupleState ts;
				_qh.decodeFVL(cachedFvl, p, len, tid, tastep, ts);
				
				_pTM->deleteDataTable(tabSetId, tableName, CegoObject::TABLE, dp, cachedFvl, cachedIdxList, cachedKeyList, false, true);
				
			    }
			    catch ( Exception e)
			    {   
				_pTM->releaseDataPtr(lockId);
				
				Chain msg;
				e.pop(msg);
				
				// _pDBMng->log(_modId, Logger::NOTICE, Chain("Incomplete rollback for table ") + tableName + Chain(" : ") + msg);
				throw Exception(EXLOC, Chain("Cannot delete entry for table"), e);
			    }
			    _pTM->releaseDataPtr(lockId);
			    
			} 
		    }
		    moreTuple = _pTM->getNextTuple(pOC, schema, dp);
		}
		else
		{
		    throw Exception(EXLOC, "Cannot get tid entry");
		}
	    }
	}
	catch ( Exception e )
	{
	    if ( pOC )
	    {
		pOC->abort();
		delete pOC;
	    }
	    throw Exception(EXLOC, Chain("Cannot rollback transaction"), e);
	}
	
	pOC->abort();
	delete pOC;
	
	// step 2 ( Purging)
	
	pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);
	
	try
	{
	    
	    CegoDataPointer dp;
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	    
	    while (moreTuple)
	    {
		
		CegoDataPointer ddp = dp; 
		moreTuple = _pTM->getNextTuple(pOC, schema, dp);
		
		_pTM->deleteData(CegoObject::RBSEG, tabSetId, ddp);
		
	    } 
	}
	catch ( Exception e )
	{
	    if ( pOC )
	    {
		pOC->abort();
		delete pOC;
	    }
	    throw Exception(EXLOC, Chain("Cannot rollback transaction"), e);
	}
	
	pOC->abort();
	delete pOC;
	
    }
    catch ( Exception e )
    {
	taLock[tabSetId].unlock();
	throw Exception(EXLOC, Chain("Cannot rollback transaction"), e);
    }

    taLock[tabSetId].unlock();
    
    return opCount;
}

void CegoTransactionManager::getOpenTid(int tabSetId, SetT<int>& tidSet)
{
    taLock[tabSetId].readLock(TS_LOCKTIMEOUT); 

    try
    {
	CegoObjectCursor *pOC;
	
	pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);
	
	try  {
	    
	    ListT<CegoField> schema = _rbcatSchema;
	    
	    CegoDataPointer dp;
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	    
	    while (moreTuple)
	    {
		
		CegoField *pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TID));
		if (pF)
		{
		    int tid;
		    memcpy(&tid, pF->getValue().getValue(), sizeof(int));
		    tidSet.Insert(tid);
		    // tidSet.Insert ( *(int*)(pF->getValue().getValue()));
		}
		moreTuple = _pTM->getNextTuple(pOC, schema, dp);
	    }
	}
	catch ( Exception e )
	{
	    pOC->abort();
	    delete pOC;
	    throw Exception(EXLOC, Chain("Cannot get open tid"), e);
	}
	
	pOC->abort();
	delete pOC;
	
	
	pOC = _pTM->getObjectCursor(tabSetId, SYS_UPDTAB, SYS_UPDTAB, CegoObject::RBSEG);
	
	try
	{
	    ListT<CegoField> schema = _updSchema;
	    
	    CegoDataPointer updp;
	    bool moreTuple = _pTM->getFirstTuple(pOC, schema, updp);
	    
	    while (moreTuple)
	    {
		
		CegoField *pF = schema.Find(CegoField(SYS_UPDTAB, SYS_UPDTAB_TID));
		
		if (pF)
		{
		    int tid;
		    memcpy(&tid, pF->getValue().getValue(), sizeof(int));
		    tidSet.Insert(tid);

		    // tidSet.Insert ( *(int*)(pF->getValue().getValue()));
		}    
		moreTuple = _pTM->getNextTuple(pOC, schema, updp);
	    }
	}
	catch ( Exception e )
	{
	    pOC->abort();
	    delete pOC;
	    throw Exception(EXLOC, Chain("Cannot get open tid"), e);
	}
	
	pOC->abort();
	delete pOC;

    }
    catch ( Exception e )
    {
	taLock[tabSetId].unlock();
	throw Exception(EXLOC, Chain("Cannot get open tid"), e);
    }

    taLock[tabSetId].unlock();
    
}

void CegoTransactionManager::recordUpdate(int tabSetId, int tid, const CegoDataPointer& dp )
{

    int fileId = dp.getFileId();
    int pageId = dp.getPageId();
    int offset = dp.getOffset();

    CegoFieldValue fv1(INT_TYPE, &tid, sizeof(int));
    CegoField f1(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_TID, INT_TYPE, sizeof(int), fv1, true, 1));	

    CegoFieldValue fv2(INT_TYPE, &fileId, sizeof(int));
    CegoField f2(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_FILEID, INT_TYPE, sizeof(int), fv2, true, 2));	

    CegoFieldValue fv3(INT_TYPE, &pageId, sizeof(int));
    CegoField f3(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_PAGEID, INT_TYPE, sizeof(int), fv3, true, 3));	

    CegoFieldValue fv4(INT_TYPE, &offset, sizeof(int));
    CegoField f4(CegoField(SYS_UPDTAB, SYS_UPDTAB, SYS_UPDTAB_OFFSET, INT_TYPE, sizeof(int), fv4, true, 4));

    ListT<CegoField> fl;
    fl.Insert(f1);
    fl.Insert(f2);
    fl.Insert(f3);
    fl.Insert(f4);

    CegoTableObject oe;
    _pTM->getObject(tabSetId, SYS_UPDTAB, CegoObject::RBSEG, oe);

    CegoDataPointer idp;
    _pTM->insertDataTable(oe, fl, idp, false);
    
}

void CegoTransactionManager::commitUpdate(int tabSetId, const Chain& idxName, int tid, bool doAppend)
{

    CegoObjectCursor *pOC = _pTM->getObjectCursor(tabSetId, SYS_UPDTAB, SYS_UPDTAB, CegoObject::RBSEG);
   
    ListT<CegoField> schema = _updSchema;
    
    CegoDataPointer updp;
    bool moreTuple = _pTM->getFirstTuple(pOC, schema, updp);
    
    while (moreTuple)
    {
	
	CegoField *pF = schema.Find(CegoField(SYS_UPDTAB, SYS_UPDTAB_TID));

	if (pF)
	{

            // if ( *(int*)pF->getValue().getValue() == tid )
	    int val;
	    memcpy(&val, pF->getValue().getValue(), sizeof(int));
	    if (val == tid)		
	    {
		
		pOC->abort();
		delete pOC;
		
		int fileId, pageId, offset;
		
		CegoField *pF = schema.Find(CegoField(SYS_UPDTAB, SYS_UPDTAB_FILEID));
		if (pF)
		    fileId = *(int*)pF->getValue().getValue();
		
		pF = schema.Find(CegoField(SYS_UPDTAB, SYS_UPDTAB_PAGEID));
		if (pF)
		    pageId = *(int*)pF->getValue().getValue();
		
		pF = schema.Find(CegoField(SYS_UPDTAB, SYS_UPDTAB_OFFSET));
		if (pF)
		    offset = *(int*)pF->getValue().getValue();
				
		CegoDataPointer dp(fileId, pageId, offset);
		
		ListT<CegoTableObject> idxList;
		
		CegoTableObject ioe;

		CegoBufferPage bp;
		_pTM->getObjectWithFix(tabSetId, idxName, CegoObject::INDEX, ioe, bp);			
		
		ListT<CegoField> fvl = ioe.getSchema();
		
		char* p;
		int len;
		
		CegoDataPointer sysEntry(bp.getFileId(), bp.getPageId(), bp.getEntryPos());
		
		unsigned long lockId = _pTM->claimDataPtr(tabSetId, CegoLockHandler::READ, CegoBufferPool::SYNC, dp, p, len);
		
		int tid;
		int tastep;
		CegoTupleState ts;
		_qh.decodeFVL(fvl, p, len, tid, tastep, ts);
		
		char ip[TABMNG_MAXINDEXVALUE];
		int ilen = TABMNG_MAXINDEXVALUE;
		int idxLen;
		CegoDataPointer ritp;
		
		_pTM->extractIndexValue(fvl, ioe.getSchema(), ip, ilen, idxLen);
		
		bool resolveIt = true;
		
		try
		{
		    CegoIndexManager idxMng(_pTM, ioe.getType());
		    idxMng.insertNativeIndexTable(ioe, dp, ip, idxLen, tid, ritp, sysEntry, doAppend);
		}
		catch ( Exception  e )
		{
		    		    
		    _pTM->releaseDataPtr(lockId);		    
		    _pTM->getDBMng()->bufferUnfix(bp, true, _pTM->getLockHandler());

		    throw Exception(EXLOC, Chain("Cannot commit update"), e);
		    
		}
		
		_pTM->releaseDataPtr(lockId);
		
		// now we can release the index root page
		_pTM->getDBMng()->bufferUnfix(bp, true, _pTM->getLockHandler());
		
		_pTM->deleteData( CegoObject::RBSEG, tabSetId, updp);
		
		pOC = _pTM->getObjectCursor(tabSetId, SYS_UPDTAB, SYS_UPDTAB, CegoObject::RBSEG);	
		
		moreTuple = _pTM->getFirstTuple(pOC, schema, updp);
		
	    }
	    else
	    {
		moreTuple = _pTM->getNextTuple(pOC, schema, updp);
	    }
	}
	else
	{
	    pOC->abort();
	    delete pOC;	    
	    throw Exception(EXLOC, "Cannot get tid entry");
	}
    }
    pOC->abort();
    delete pOC;
    
}

void CegoTransactionManager::getTransactionInfo(int tabSetId, int tid, int& undoop, int& updop)
{
    
    CegoObjectCursor *pOC;
    
    pOC = _pTM->getObjectCursor(tabSetId, SYS_RBCATLOG, SYS_RBCATLOG, CegoObject::RBSEG);

    try  {
	
	ListT<CegoField> schema = _rbcatSchema;
	
	CegoDataPointer dp;
	bool moreTuple = _pTM->getFirstTuple(pOC, schema, dp);
	
	undoop=0;
	while (moreTuple)
	{
	    
	    CegoField *pF = schema.Find(CegoField(SYS_RBCATLOG, SYS_RBCATLOG_TID));
	    if (pF)
	    {
		// if (*(int*)(pF->getValue().getValue()) == tid)

		int val;
		memcpy(&val, pF->getValue().getValue(), sizeof(int));
		if (val == tid)		
		{
		    undoop++;
		}
	    }
	    moreTuple = _pTM->getNextTuple(pOC, schema, dp);
	}
    }
    catch ( Exception e )
    {
	pOC->abort();
	delete pOC;
	throw Exception(EXLOC, Chain("Cannot get transaction info"), e);
    }
    
    pOC->abort();
    delete pOC;


    pOC = _pTM->getObjectCursor(tabSetId, SYS_UPDTAB, SYS_UPDTAB, CegoObject::RBSEG);
   
    try
    {
	ListT<CegoField> schema = _updSchema;
    
	CegoDataPointer updp;
	bool moreTuple = _pTM->getFirstTuple(pOC, schema, updp);
	
	updop=0;
	while (moreTuple)
	{
	    
	    CegoField *pF = schema.Find(CegoField(SYS_UPDTAB, SYS_UPDTAB_TID));
	    
	    if (pF)
	    {

		// if ( *(int*)pF->getValue().getValue() == tid )

		int val;
		memcpy(&val, pF->getValue().getValue(), sizeof(int));
		if (val == tid)		
		{	
		    updop++;
		}
	    }
	    
	    moreTuple = _pTM->getNextTuple(pOC, schema, updp);
	    
	}
    }
    catch ( Exception e )
    {
	pOC->abort();
	delete pOC;
	throw Exception(EXLOC, Chain("Cannot get transaction info"), e);
    }
    
    pOC->abort();
    delete pOC;
       
}
